/* -*-C-*- lptops.c */
/*-->lptops*/
/**********************************************************************/
/******************************* lptops *******************************/
/**********************************************************************/

/***********************************************************************
	       Line printer file to PostScript converter

lptops  converts normal text  files  into PostScript  for display on any
PostScript-compatible output device.

lptops  converts normal text files  into PostScript  for  printing  on a
PostScript  printer  or typesetter,  or any  other PostScript-compatible
output device.  It supports:

	*   selection of a variety of fonts at arbitrary point sizes
	    with adjustable interline spacing (e.g. for double spacing),

	*   margin specification,

	*   page outlining,

	*   portrait and landscape page orientation,

	*   automatic page numbering,

	*   multi-column printing,

	*   overstruck text (used by many text  formatters  for
	    underlining and bolding),

	*   UNIX manual page output from nroff(1),

	*   ISO8859 Latin-1 (an 8-bit superset of ASCII) and Adobe
	    PostScript character set encodings,

	*   high-precision character coordinates  for  accurate
	    typesetter output,

	*   initialization file for local customization,

	*   downloadable PostScript fonts,

	*   mapping of PostScript font names to system-dependent file
	    names,

	*   full conformance of output PostScript to Adobe Document
	    Structuring Conventions version 3.0 (see Appendix G of the
	    Adobe red book, PostScript Language Reference Manual, second
	    edition, Addison-Wesley (1990), ISBN 0-201-1812704),

	*   clipping of long text lines, and

	*   wrapping of long text lines.

lptops  can be  compiled with  a wide variety of  C and C++ compilers on
most common operating systems.  The source  code distribution for lptops
is in the  PUBLIC DOMAIN.  The author will be  grateful for bug reports,
bug fixes, extensions, and news of wider use of this program.

A complete description can be found in the associated manual page files,
lptops.man and lptops.hlp.

The following preprocessor variables can be defined at compile time:

AFMENVVAR		Adobe font metric file search path variable
AFMEXT			Adobe font metric file extension
AFMMAPCOMMENTCHAR	Adobe font mapping file comment starter
AFMMAPFILE		Adobe font mapping file name
AFMMAPSEPCHARS		Adobe font mapping file token separator characters
AFMPATH			Adobe font metric file search path
FONTSEPCHARS		font name separators in -f=font1,font2,... option
INITENVVAR		initialization file environment variable name
INITFILE		initialization file name
INITFILECOMMENTCHAR	initialization file comment starter
INITFILESEPCHARS	initialization file token separator characters
INITPATH		initialization file search path
MAXEXTRAFONTS		limit on extra run-time-defined fonts
MAXEXTRAPAPER		limit on extra run-time-defined paper sizes
OPTIONCHARS		command-line option starter character(s)
PAPERSEPCHARS		separator characters in -paper=name,width,height
PAPER_DEFAULT		index in paper[] of default paper type

The source code below gives default values.  Most of these should NOT be
changed, unless the documentation is updated.  Choice of a default paper
type is best made in an initialization file.

In order to handle overprinting of input of the form

...text...<CR>...overtext...<CR><LF>

lptops must be  loaded with  a  special version  of  the  C library  I/O
routine (PCC:IOCR.REL  on the DEC-20) which  permits reading and writing
of <CR>.  No special action need be taken on UNIX systems, since <CR> is
not filtered  out, and is normally absent.	On PC-DOS, open options
of  "rb" will usually  be necessary (this is the default on all  systems
but TOPS-20).

Because  of  the variety  in run-time  library  mapping  of  end-of-line
designators to C's (and UNIX's)  '\n', output  of  newline characters is
relegated  to  NEWLINE() and  OUTNL()  macros, and  strictly  avoided in
strings.

========================================================================

Revision history (reverse time order):

[23-Jan-2004]	Version 3.1.8
		Minor update to ensure that cuserid() always has a
		prototype.  Otherwise, on systems like Intel Itanium-2
		where pointers by default exceed 2^32 (e.g.,
		0x20000000002bf870), truncating a pointer to a default
		int will produce an address violation.  This bug was
		not caught years ago on 64-bit DEC Alpha OSF/1 or Sun
		SPARC Solaris because cuserid() has a prototype there,
		nor on SGI MIPS IRIX, because pointers normally fit in
		32 bits in a 64-bit world.  However, on Red Hat Linux
		Advanced Server release 2.1AS (Derry) on Intel
		Itanium-2, the cuserid() prototype is not exposed
		unless compilation is done with _XOPEN_SOURCE defined.

		Add code to use the more secure mkstemp() when
		available, in place of tmpnam(), which is subject to
		guessing attacks.  Temporary file opens now go
		through the new function Fopen_Tempfile().

[02-Jan-2003]	Version 3.1.7
		Update documentation lptops.man to describe page status
		reports, add new AVAILABILITY section, and revise
		COPYRIGHT section since this version is now placed under
		the GNU General Public License.

		Update version number and date in configure.in and
		lptops.c, and regenerate configure with latest autoconf
		and autoheader.

		Replace copyright[] message in Opt_Copyright() to
		reflect license change.

[23-Nov-2002]	Version 3.1.6
		Correct output PostScript header comment to avoid 
		conflict with UNIX CUPS line printer spooler.

[28-Aug-1997]	Version 3.1.5
		Prepare standalone version that uses GNU automake and
		autoconf to prepare a Makefile and configure script,
		allowing simple UNIX installation like this:

			./configure && make all check install

		The run-time behavior of lptops remains identical to
		that at the previous version.

[28-Oct-1996]	Version 3.1.4
		Correct %%DocumentMedia comment; fourth parameter should
		be zero instead of empty string.

		Add -?,  -copyleft,  and -help options.   Make them, and
		-author,  -copyright,  and  -version  exit  successfully
		after the requested information has been displayed.

		Update author information to include World-Wide Web URL.

[07-May-1996]	Version 3.1.3
		Add 36 new U.S. paper types found in an article in
		Adobe Magazine.  Update man pages, and create a standard
		file header in lptops.man.

[12-Jul-1995]	Version 3.1.2
		Replace sscanf() by tscanf(),  thereby avoiding bugs  in
		several vendor-provided sscanf()  implementations,   and
		guaranteeing support for character set ranges.

[02-Nov-1994]	Version 3.1.1
		Correct definition of  OUTSTR() so that cast is  correct
		if the argument  is an  expression.  Change basename  to
		the_basename  in  Download_Font()  to avoid  shadowing a
		global symbol on some systems.

[26-Feb-1994]	Version 3.1.0
		Add -page  switch to permit   selection of output pages.
		This proves useful  for restarting print jobs  that have
		jammed  in the printer, or  for printing alternate pages
		for two-pass duplex  printing.  The new functions  added
		are Check_Page_Selection() and Opt_SelectPages(), plus a
		new typedef PAGE_SELECTION.

[14-Aug-1993]	Version 3.0.2
		Fix error in  Out_Vskip() that resulted in output tokens
		not being   space-separated.   Thanks to  John Jorgensen
		<jorgnsn@qucis.queensu.ca> for reporting this bug.

		Extend Copy_Plist() to   allow tab characters  in Scribe
		@X[...] font change sequences.

[10-Jun-1993]	Version 3.0.1
		Extend  Reverse_Pages()  to  handle  possible failure of
		ftell(stdout); this happens on HP-UX  9.0 when stdout is
		directed  to  a  pipe.   TEMPFILE  redefined to  support
		multiple temporary  files.  Add  workarounds for bug  in
		HP-UX  9.0  that  results in erroneous fprintf()  return
		values when ftell() has been called.

		Extend Option()  to  allow (possibly-abbreviated) values
		yes, no, true, and false for Boolean options.

		Make ps_xxx[] tables static to reduce global names.

[08-Apr-1993]--[29-Apr-1993]	Version 3.0.0
		Add -clip option, and output the true  bounding box when
		clipping is in effect.

		Add  the QSU_TO_xx(), INCH_TO_QSU(), and  DIMEN_TO_QSU()
		macros  to  simplify  and  clarify  conversions  between
		various units.

		Increase  the  value  of  QSU,  the   output  precision.
		Comments at the point of definition explain why.

		Add COLUMNPATH and NEWCOLUMN macros, and move drawing of
		optional  outline box  from  End_Col() to Beg_Col(),  so
		that it can be drawn before an optional clipping path is
		set.

		Add Out_xxx() functions to greatly simplify Do_File().

		Add Out_Lines() function to allow table-driven output in
		many parts of the program.

		Add Usage() function.

		Add Dump_Font() and  Font_Option() functions to simplify
		coding.

		Add support  for reading Adobe Font  Metric files.   AFM
		files are searched for in a compiled-in default list  of
		directories which  may be overridden by a  path list  in
		the  AFMPATH environment variable.   Add  -e option  for
		changing character set from ISO8859-1 Latin 1 (ASCII) to
		PostScript.  New  functions are Do_ISO8859_1_encoding(),
		Read_Metrics(), and  Reencode().  This  is a significant
		new capability,  so  the major version  number  has been
		increased  from  2  to  3.    Thanks   to   Sam  Leffler
		<sam@sgi.com>, who implemented  a preliminary version of
		this  in  an  older  version  of  lptops in his  flexfax
		package.   While  the general features of  his additions
		have  been  incorporated  in this  version, many changes
		were made to simplify and clarify the code.

		Add Map_Font()  to handle the mapping of long PostScript
		font names to system-dependent file names.  The  mapping
		is provided by an external file  named by the AFMMAPFILE
		preprocessor symbol (default: afmfonts.map) found in the
		AFMPATH search  path.   It contains lines  with pairs of
		font   names   and  local   pathless,   and  optionally,
		extensionless, file names,  separated  by AFMMAPSEPCHARS
		(default:  space,  tab,  equals,  and colon).   Optional
		comments run from AFMMAPCOMMENTCHAR  (default:  percent)
		to end of line;  a backslash  preceding such a character
		removes its significance as a comment starter.  A subset
		of this  syntax  is used  by Adobe's  TranScript package
		file, afmfiles.upr, and we have used it to provide short
		file  names  in the  afmfonts.map  file that accompanies
		this software.

		Add  -wrap option.   This uses character widths for  the
		normal font from the  .afm  file, and  does not yet take
		account of Scribe-style font changes to bold and italic;
		the metrics of the normal font are used  instead.   This
		omission should rarely be noticeable.

		Add -manpage option.  This causes underline sequences of
		<underline><backspace><char>  to underline  the  <char>.
		UNIX manual page output from  nroff  uses this  encoding
		extensively.   Also recognized are <o><backspace><+> and
		<+><backspace><o>  for  a  bullet, <ESC><7>  for reverse
		line  feed, <ESC><8>  for reverse  half  line feed,  and
		<ESC><9> and <VT> for reverse line feed.  New functions:
		Out_Escape(), Out_Manchar(), and Out_Vskip().

		Add GETC()  and ungetchar() for -manpage option support.
		Standard C guarantees  only  one  character of pushback,
		but we need at least 3.

		Add   Get_Bold_Char()   and   Get_Underline_Char()   for
		-manpage option support.   These  support typesetting of
		multi-character  bold  or  underlined  strings  by   one
		command.

		Add  Do_Scribe_string()  to  simplify  the interface  to
		Copy_Plist().

		Add Scribe-like @U[...] for underlined text, now that we
		have general support for it through the U macro.

		Add Scribe-like @T[...] for typewriter text, and @H[...]
		for bold-italic or bold-oblique text.

		Modify macros  so  that a  single macro now  selects the
		font and  prints  a string, instead of  having  separate
		macros.  This reduces  the  output  volume,  and permits
		clean support of underlined fonts.

		Add Fatal() and Warning() to centralize message control,
		and shorten  coding.   Call  perror()  inside  them when
		errno is non-zero, and zero errno  at key code points to
		avoid bogus calls to perror().

		Add initialization file  support and  -f=font1,...,font5
		option  for  defining  additional fonts, possibly  to be
		downloaded   from   host  .pfa  or  .pfb   formats.   An
		initialization   file  can  contain   any   command-line
		options, and comments from INITFILECOMMENTCHAR (default:
		%)  to end of  line.   It  is read  and processed before
		command-line  arguments   so  it  can   provide  locally
		customized defaults that the user can still override.

		Because  options  can  now be set in  an  initialization
		file, it is necessary to have a way  to override Boolean
		options.  All such options now support the form  -option
		to toggle the current setting, -option:<zero> to turn it
		off, and -option:<non-zero> to turn it on.

		Add    Do_Startup_File(),    Get_Wrapped_Line(),     and
		Strip_Comments() functions.

		Add  OPTIONCHARS  to  allow  the compile-time setting of
		command-line option  prefix  characters  to  distinguish
		options from file names.

		Add  Font typedef,  and  extend Read_Metrics() to take a
		Font argument so as to permit separate font metrics  for
		normal, bold, italic, bold-italic, and typewriter fonts.
		Add Set_Family() to simplify setting fonts in a family.

		Add support for downloaded PostScript fonts in .pfa  and
		.pfb formats, with new functions Copy_PFA(), Copy_PFB(),
		Copy_PFB_Binary(), Copy_PFB_Text(), Download_Font(), and
		PFB_Length().   Extend Map_Font() to allow  setting  the
		file extension, so that the extension of  the  file name
		in  the  font mapping file can be  changed from  .afm to
		.pfa or .pfb, as needed.  The lptops source distribution
		now  includes  Adobe  Utopia,  Bitstream  Charter,   and
		Bitstream Courier fonts in .afm, .pfa, and .pfb formats.
		These  fonts,   although  copyrighted,  may   be  freely
		redistributed.  They were contributed  to  the  X Window
		Consortium by Adobe and Bitstream.

		Add   %%DocumentFonts,   %%DocumentNeededResources,  and
		%%DocumentSuppliedResources  comments.    And   function
		Do_Resources() to  record  font usage  in  the  document
		trailer.

		Wrap the  %%Title and %%PageTable output  lines  to keep
		line lengths modest.

		Tighten  the  error checking in Option()  to reject  all
		misspelled options and bad values.

		Add  support  for  multiple paper  sizes,  with built-in
		table in  paper[],  -paper  option,  and  new  functions
		Dump_Paper() and Opt_Paper().  Default paper is "A", but
		that  can  be  changed  by  a  compile-time  setting  of
		PAPER_DEFAULT, or better, a  -paper:xxx  setting  in the
		initialization file.  Add %%DocumentMedia comment.

		Add support  for line wrapping of long output PostScript
		lines with  new  functions  Out_Char(), Out_Need(),  and
		Out_StrChar().   Adobe  guidelines  require   that  line
		lengths not  exceed 255 characters,  and 80 is  an  even
		better  limit for portability.  lptops  must be  able to
		correctly handle input lines of arbitrary length, so  it
		must also be able  to  wrap them on output.   Because of
		PostScript's idiosyncratic line continuation rules, this
		requires  tracking  of  the  output  state  (normal,  in
		comment,  or  in  parenthesized  string).   Because  the
		lptops comment lines have  a  well-defined form, and are
		not expected to be long except in the pathological  case
		of an extremely  long title,  we handle line wrapping of
		long  titles  separately, and thereby avoid  needing  to
		continue comments.

		Create a set of input files  for a torture test, to test
		and validate new installations.

		Add Do_PageTable(), Do_Trailer(), and Do_Validation().

		Add  typedef  OPTION, T_xxx flags,  Option_Table[],  and
		make Option() entirely table-driven.

		Extend several option names to multi-character names for
		readability, allowing the original  one-letter forms  to
		be  recognized.  Replace  old -a with -scribe,  -u  with
		-reverse, and  -h  with -landscape;  the originals  were
		unmnemonic  abberations  caused  by  the  old one-letter
		option name restriction.  Add -copyright option.  Remove
		version number and copyright output from -author output.
		Remove copyright output from -version output.

		Change default page order to non-reversed, since this is
		appropriate for most newer printers.  The old default of
		page reversal can be obtained by  an appropriate setting
		in an initialization file.

[06-Mar-1993]	Version 2.2.6
		Update author telephone and FAX numbers.

[08-Dec-1992]	Version 2.2.5
		Revise Do_File()  and Copy_Plist()  to map  @@  to  @ in
		Scribe  mode, so that @  can appear in normal, bold, and
		italic text.

[02-Jul-1992]	Version 2.2.4
		Revise Do_File() and inchar() further to avoid output of
		empty page  when the  last page  is  filled  completely.
		Berkeley  pr ends  its output  with <newline><formfeed>,
		which version  2.2.2 handled.   AT&T  pr ends its output
		with <formfeed>, which this version handles.

[16-Apr-1992]	Version 2.2.3
		Split  body  of  main() into  smaller  functions.   Make
		internal  functions  static.   Output  long  strings  in
		pieces to avoid  line-wrapped constant strings in source
		code.

[03-Jan-1992]	Version 2.2.2
		Revise Do_File() to avoid  output of empty page when the
		last page is filled completely; this saves paper, at the
		cost of	  two extra  tests  for every  input  character.
		Revise inchar()  to  handle  final  formfeed  specially.
		Also, correct off-by-one line number status reports.

[07-Jun-1991]	Version 2.2.1
		Fix landscape output  problem;  Do_Prolog()  incorrectly
		preceded setting of page_height and page_width.

[06-Jun-1991]	Version 2.2.0
		Add -q (quiet) option.

[31-May-1991]	Version 2.1.0
		Update for EPSF 3.0 conformance.   This includes several
		PostScript   comment changes.   Previously, the showpage
		command  was embedded inside  a save/restore, and  there
		was a  global save/restore  around the whole job.  These
		interfered with  redefinition of  showpage.   The global
		pair has been removed.  The /#copies is now set  only in
		the setup section, instead  of in every page,  following
		Adobe's     EPSF  guidelines.    Control-D   end-of-file
		terminator   has   been  replaced   by  %%EOF   comment.
		Formfeeds at page ends are suppressed unless FORMFEED is
		defined; Adobe recommends  not using formfeeds, so  this
		is off by default.

[01-Apr-1991]	Add -author, -version, and -y# options,   labelling this
		new  version  2.0.0.    Copyright  information  is  also
		displayed for -version.

[03-May-1988]	Add ditroff and Adobe TranScript font abbreviations.

[19-Apr-1986]	Change page/line status report from output in End_Page()
		only to output in Beg_Page() and End_Column().

[07-Apr-1986]	Fix some  incorrect  font  names, put  font  names  into
		table, add font abbreviations and new font names, change
		default font to Courier  (because most printing  expects
		fixed-width fonts).  Add usage message  if  bad options,
		and print  font names  if  a bad  font option is  given.
		Allow multiple filenames on command line instead of just
		using  stdin---this  added  new  function  Do_File() and
		changed  argument lists  of inchar()  and  Copy_Plist().
		Add save/restore around each page  image  following  new
		1986   Adobe   PostScript  Language   Reference   Manual
		recommendation, and change multiple copies sequence from
		repeat  loop to setting  of  standard  variable #copies.
		Add -U option for unsorted output.

[03-Feb-1986]	Fix setting of final  entry of page_table[].  It  should
		only be done when End_Page() is called, since  otherwise
		Beg_Page() has already set the final entry, and we  just
		have an empty page to  be discarded.  Without this  fix,
		this left a hole (0), so  the entire file got copied  in
		its entirety to the last page of the sorted file.

[18-Jan-1986]	Add (void) to fprintf()/printf(), test for disk  storage
		full at end  of page, add  OUTSTR() macro using  fputs()
		instead of fprintf() for efficiency.

[20-Dec-1985]	Change	getchar() to inchar();	the  latter  supplies  a
		trailing    LF	at EOF	in  the	   event it is	missing.
		Otherwise, the output PostScript  has two tokens  jammed
		together (P 1 {copypage} --> P1 {copypage})

[20-Dec-1985]	Following recommendation of Allan Hetzel on LASER-LOVERS
		ARPA BBOARD, added "() pop" at top of every page.   This
		is selected by the preprocessor variable FIX31OCT85:

		--------------------------------------------------------
		From: Allan Hetzel  <SYSAL%UKCC.BITNET@WISCVM.ARPA>
		Subject: Re: Apple LaserWriter XON/XOFF problem
		To:  <LASER-LOVERS@WASHINGTON.ARPA>

		The note  posted to  Laser-Lovers on  Oct 28  by  Joseph
		Goldstone of  Symbolics  Cambridge Research  Center  was
		helpful but I didn't try his suggested fix correctly and
		so I called Adobe.  They were very helpful (much more so
		than Apple) and explained the problem and how to  bypass
		it.

		My  apologies  to   Adobe  if  this   not  an	accurate
		description of the problem.   The problem apparently  is
		due to the  PostScript interpreter  getting confused  by
		what it thinks is an excess of white space characters.

		The bypass is to place a special character (parenthesis,
		bracket, brace, slash, etc.) to the right of some token,
		but without an intervening white space character. As the
		PostScript scanner  reads the  token it  also reads  the
		special character which acts as a delimiter. The scanner
		then has  to  back  up  one character  so  that  it  can
		reprocess the special character  after it has  processed
		the preceding token.  During  this backing up process  a
		number of  internal values  are recalculated,  including
		some pointer  into the  input buffer.	This causes  the
		XON/XOFF protocol to work properly.  Doing this once per
		page seems to keep everybody happy.

		Each page  in  the PostScript  file  built by  our  word
		processing program  is surrounded  by a  "save  restore"
		sequence.  We  changed	 the  beginning  save	sequence
		"/saveobj save def" to  read "/saveobj save def()  pop".
		The "() pop" is effectively  a no-op and we are  assured
		that the necessary recalculations are done on each page.
		This seems to have corrected the problem completely.

		Allan Hetzel (sysal@ukcc.bitnet)
		--------------------------------------------------------

[11-Dec-1985]	Fix problem with extra () around overstruck text.

[13-Nov-1985]	Insert FASTIO conditional compilation option to speed up
		page reversal.  Under VAX  4.2BSD, fread() and  fwrite()
		are already fast enough, and need not be replaced.

[12-Nov-1985]	Add reversal of  pages, storing  output on  intermediate
		file which is deleted before exiting.

[02-Nov-1985]   Add  code  to  write  output  on  temporary  file,  then
		retrieve in reverse page order and output on stdout.

[30-Oct-1985]	Add code to handle overstruck lines:
		...text...<CR>...overtext...<CR><LF>
		will appear as
		(...text...) O
		(...overtext...) P

		Macro definitions and font selection are now repeated on
		each page to allow selective reprinting.

[07-Oct-1985]	Add code to discard input NUL's  from word-padded files,
		since the LaserWriter prints them as blanks.

[27-Sep-1985]	Added   missing  'N'  option,  fixed  Page_Option()  (it
		scanned beyond last digit of page number)

[11-Jul-1985]	Original version.

***********************************************************************/

#include "headers.h"

#undef VERSION
#define VERSION \
    "lptops [Line Printer to PostScript] version 3.1.7 [02-Jan-2003]"

#if 0
#define VERSION \
    "lptops [Line Printer to PostScript] version 3.1.6 [23-Nov-2002]"

#define VERSION \
    "lptops [Line Printer to PostScript] version 3.1.5 [29-Aug-1997]"

#define VERSION \
    "lptops [Line Printer to PostScript] version 3.1.4 [28-Oct-1996]"

#define VERSION \
    "lptops [Line Printer to PostScript] version 3.1.3 [07-May-1996]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 3.1.2 [12-Jul-1995]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 3.1.1 [02-Nov-1994]"

#define VERSION \
    "lptops [Line Printer to PostScript] version 3.1.0 [26-Feb-1994]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 3.0.2 [14-Aug-1993]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 3.0.1 [03-Jun-1993]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 3.0.0 [29-Apr-1993]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 2.2.6 [06-Mar-1993]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 2.2.5 [08-Dec-1992]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 2.2.4 [02-Jul-1992]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 2.2.3 [16-Apr-1992]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 2.2.2 [03-Jan-1992]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 2.2.1 [07-Jun-1991]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 2.2.0 [06-Jun-1991]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 2.1.0 [31-May-1991]"
#define VERSION \
    "lptops [Line Printer to PostScript] version 2.0.0 [01-Apr-1991]"
#endif /* 0 */

RCSID("$Id: lptops.c,v 1.3 2004/01/23 21:51:03 beebe Exp beebe $")

#define DEFAULT_TMPFILE	"/tmp/lptops.XXXXXX"

#if    (OS_ATARI || PCC_20)
#define DISKFULL(fp)	ferror(fp)	/* PCC-20 does not always set errno */
#else  /* NOT (OS_ATARI || PCC_20) */
#if defined(ENOSPC)
#define DISKFULL(fp)	(ferror(fp) && (errno == ENOSPC))
#else
#define DISKFULL(fp)	(ferror(fp) && (errno != 0))
#endif
#endif /* (OS_ATARI || PCC_20) */

/*
************************************************************************
On  TOPS-20,   fread()/fwrite()   are  implemented   as   loops   around
getc()/putc() calls, and these in turn also do CR/LF translation.  Calls
to _read() and _write() translate into  direct calls to the Monitor  for
fast efficient block I/O.   FASTIO selects this option.   It is used  in
Copy_Block when the temporary  file is output in  reverse page order  to
stdout.
************************************************************************
*/

#if    OS_TOPS20
#if    PCC_20
#define FASTIO 1			/* TOPS-20 use only */

#ifdef DISKFULL
#undef DISKFULL
#endif /* DISKFULL */

#define DISKFULL(fp)	ferror(fp)	/* PCC-20 does not always set errno */

#endif /* PCC_20 */
#endif /* OS_TOPS20 */

#define FIX31OCT85 1			/* Allan Hetzel's XON/XOFF fix */

#ifdef CNULL				/* IBM 3090 and PS/2 define this */
#undef CNULL
#endif
#define CNULL	(const char*)NULL

#ifdef FNULL
#undef FNULL
#endif
#if defined(__sgi) && !defined(__cplusplus)
#define FNULL	(void (*) ())NULL
#else
#if 0
    /* This is Standard C */
#define FNULL	(void (*) ARGS((const char* option_,const char *s_)))NULL
#else
    /* This is for brain-damaged sub-Standard C compilers that cannot
       handle variable names in function prototypes: DEC Ultrix 4.3,
       and possibly others, sigh... */
#define FNULL	(void (*) ARGS((const char *,const char *)))NULL
#endif
#endif

#ifdef VNULL
#undef VNULL
#endif
#define VNULL	(VOIDP)NULL

typedef int BOOLEAN;

#ifdef FALSE
#undef FALSE
#endif /* FALSE */

#define FALSE 0

#ifdef TRUE
#undef TRUE
#endif /* TRUE */

#define TRUE 1

typedef long COORDINATE;	/* variables in QSU need >=32 bits (see below)*/

#define MAXFONTSIZE 256			/* number of chars per font */

typedef struct
{
    const char *name;			/* full PostScript font name */
    BOOLEAN resident;			/* TRUE if printer-resident */
    COORDINATE CharWidth[MAXFONTSIZE];	/* character width metrics */
} Font;

#define CHARWIDTH(c)	current_font->CharWidth[(unsigned int)(0xff & (c))]

typedef struct
{
    int		begin;
    int		end;
    int		step;
} PAGE_SELECTION;

typedef struct
{
    const char *name;
    const char *width;
    const char *height;
    const char *left_margin;
    const char *right_margin;
    const char *bottom_margin;
    const char *top_margin;
} Paper;

#ifndef MAXEXTRAPAPER
#define MAXEXTRAPAPER	25	/* number of extra run-time paper definitions */
#endif

#define MAXPAPER	(107 + 1 + MAXEXTRAPAPER)

#ifndef PAPER_DEFAULT
#define PAPER_DEFAULT	0		/* index of default paper type */
#endif

static Paper paper[MAXPAPER] =
{ /* this list is extended at run time by -paper:name,width,height options */

    /* name             width    height    left    right   bottom  top */

    /* ANSI (note that B folded in half becomes A, etc.) (portrait) */

    { "A",		"8.5in", "11in",   "1in",  "1in",  "1in",  "1in" },
    { "B",		"11in",  "17in",   "1in",  "1in",  "1in",  "1in" },
    { "C",		"17in",  "22in",   "1in",  "1in",  "1in",  "1in" },
    { "D",		"22in",  "34in",   "1in",  "1in",  "1in",  "1in" },
    { "E",		"34in",  "44in",   "1in",  "1in",  "1in",  "1in" },

    /* Common US designations (portrait) */

    { "Computer-1411",	"14in",	 "11in",   "1in",  "1in",  "1in",  "1in" },
    { "Legal",		"8.5in", "13in",   "1in",  "1in",  "1in",  "1in" },
    { "Letter",		"8.5in", "11in",   "1in",  "1in",  "1in",  "1in" },
    { "US-Legal",	"8.5in", "14in",   "1in",  "1in",  "1in",  "1in" },

    /* Common US designations (landscape) */

    { "A-L",		"11in",	 "8.5in",  "1in",  "1in",  "1in",  "1in" },
    { "Computer-1411-L","11in",	 "14in",   "1in",  "1in",  "1in",  "1in" },
    { "Legal-L",	"13in",	 "8.5in",  "1in",  "1in",  "1in",  "1in" },
    { "Letter-L",	"11in",	 "8.5in",  "1in",  "1in",  "1in",  "1in" },
    { "US-Legal-L",	"14in",	 "8.5in",  "1in",  "1in",  "1in",  "1in" },

    /* Additional paper and envelope sizes taken from ``Hewlett-Packard
       LaserJet Family Paper Specification Guide'' (1992). */

    { "COM10",		"4.1in", "9.5in",  "1in",  "1in",  "1in",  "1in" },
    { "DL",		"110mm", "220mm",  "25mm", "25mm", "25mm", "25mm" },
    { "Executive",	"7.25in","10.5in", "1in",  "1in",  "1in",  "1in" },
    { "Monarch",	"3.9in", "7.5in",  "1in",  "1in",  "1in",  "1in" },

    /* Miscellaneous and Imagen Ultrascript paper sizes */

    { "A4Small",	"210mm", "297mm",  "25mm", "25mm", "25mm", "25mm" },
    { "Ledger",		"11in",	 "17in",   "1in",  "1in",  "1in",  "1in" },
    { "LetterSmall",	"8.5in", "11in",   "1in",  "1in",  "1in",  "1in" },
    { "Note",		"8.5in", "11in",   "1in",  "1in",  "1in",  "1in" },

    /* Additional (U.S.) paper sizes taken from

	    Constance J. Sidles
	    On Paper: Paper fits (and what to do about them)
	    Adobe Magazine
	    p. 23--23
	    May/June 1996

       This article discusses the problems of selecting paper sizes
       that conform to industry standards, and avoid waste during
       trim operations.

       The paper sizes in this article are given only by their
       dimensions in inches, rather than by alphabetic names, so we
       simply use PAPER-TYPE:WIDTHxHEIGHT as the paper name.
    */

    /* Paper type: Bond */
    { "Bond:8.5x11",		  "8.5in",  "11in",  "1in",  "1in",  "1in",  "1in" },
    { "Bond:8.5x14",		  "8.5in",  "14in",  "1in",  "1in",  "1in",  "1in" },
    { "Bond:11x17",		  "11in",   "17in",  "1in",  "1in",  "1in",  "1in" },
    { "Bond:17x22",		  "17in",   "22in",  "1in",  "1in",  "1in",  "1in" },
    { "Bond:17x28",		  "17in",   "28in",  "1in",  "1in",  "1in",  "1in" },
    { "Bond:19x24",		  "19in",   "24in",  "1in",  "1in",  "1in",  "1in" },
    { "Bond:19x28",		  "19in",   "28in",  "1in",  "1in",  "1in",  "1in" },
    { "Bond:22x34",		  "22in",   "34in",  "1in",  "1in",  "1in",  "1in" },

    /* Paper type: Uncoated Book */
    { "UncoatedBook:17.5x22.5",	  "17.5in", "22.5in","1in",  "1in",  "1in",  "1in" },
    { "UncoatedBook:19x25",	  "19in",   "25in",  "1in",  "1in",  "1in",  "1in" },
    { "UncoatedBook:23x29",	  "23in",   "29in",  "1in",  "1in",  "1in",  "1in" },
    { "UncoatedBook:23x35",	  "23in",   "35in",  "1in",  "1in",  "1in",  "1in" },
    { "UncoatedBook:25x38",	  "25in",   "38in",  "1in",  "1in",  "1in",  "1in" },
    { "UncoatedBook:35x45",	  "35in",   "45in",  "1in",  "1in",  "1in",  "1in" },
    { "UncoatedBook:38x50",	  "38in",   "50in",  "1in",  "1in",  "1in",  "1in" },

    /* Paper type: Text */
    { "Text:17.5x22.5",		  "17.5in", "22.5in","1in",  "1in",  "1in",  "1in" },
    { "Text:23x35",		  "23in",   "35in",  "1in",  "1in",  "1in",  "1in" },
    { "Text:25x38",		  "25in",   "38in",  "1in",  "1in",  "1in",  "1in" },
    { "Text:26x40",		  "26in",   "40in",  "1in",  "1in",  "1in",  "1in" },

    /* Paper type: Coated Book */
    { "CoatedBook:19x25",	  "19in",   "25in",  "1in",  "1in",  "1in",  "1in" },
    { "CoatedBook:23x29",	  "23in",   "29in",  "1in",  "1in",  "1in",  "1in" },
    { "CoatedBook:23x35",	  "23in",   "35in",  "1in",  "1in",  "1in",  "1in" },
    { "CoatedBook:25x38",	  "25in",   "38in",  "1in",  "1in",  "1in",  "1in" },
    { "CoatedBook:35x45",	  "35in",   "45in",  "1in",  "1in",  "1in",  "1in" },
    { "CoatedBook:38x50",	  "38in",   "50in",  "1in",  "1in",  "1in",  "1in" },

    /* Paper type: Cover */
    { "Cover:20x26",		  "20in",   "26in",  "1in",  "1in",  "1in",  "1in" },
    { "Cover:23x35",		  "23in",   "35in",  "1in",  "1in",  "1in",  "1in" },
    { "Cover:25x38",		  "25in",   "38in",  "1in",  "1in",  "1in",  "1in" },
    { "Cover:26x40",		  "26in",   "40in",  "1in",  "1in",  "1in",  "1in" },

    /* Paper type: Index or Board */
    { "Index:22x28",		  "22in",   "28in",  "1in",  "1in",  "1in",  "1in" },
    { "Index:22.5x28.5",	  "22.5in", "28.5in","1in",  "1in",  "1in",  "1in" },
    { "Index:23x29",		  "23in",   "29in",  "1in",  "1in",  "1in",  "1in" },
    { "Index:23x35",		  "23in",   "35in",  "1in",  "1in",  "1in",  "1in" },
    { "Index:24x36",		  "24in",   "36in",  "1in",  "1in",  "1in",  "1in" },
    { "Index:25.5x30.5",	  "25.5in", "30.5in","1in",  "1in",  "1in",  "1in" },
    { "Index:28x44",		  "28in",   "44in",  "1in",  "1in",  "1in",  "1in" },

    /* ISO standard paper A-series (portrait).   Note that Xn folded in half
       becomes Xn+1 (X = A, B, C) */

    { "A0",		"841mm", "1189mm", "25mm", "25mm", "25mm", "25mm" },
    { "A1",		"594mm", "841mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A2",		"420mm", "594mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A3",		"297mm", "420mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A4",		"210mm", "297mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A5",		"148mm", "210mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A6",		"105mm", "148mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A7",		"74mm",  "105mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A8",		"52mm",  "74mm",   "25mm", "25mm", "25mm", "25mm" },
    { "A9",		"37mm",  "52mm",   "25mm", "25mm", "25mm", "25mm" },
    { "A10",		"26mm",  "37mm",   "25mm", "25mm", "25mm", "25mm" },

    /* ISO standard paper A-series (landscape).  Note that the final L
       is NOT prefixed by a hyphen. */

    { "A0L",		"1189mm","841mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A1L",		"841mm", "594mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A2L",		"594mm", "420mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A3L",		"420mm", "297mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A4L",		"297mm", "210mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A5L",		"210mm", "148mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A6L",		"148mm", "105mm",  "25mm", "25mm", "25mm", "25mm" },
    { "A7L",		"105mm", "74mm",   "25mm", "25mm", "25mm", "25mm" },
    { "A8L",		"74mm",  "52mm",   "25mm", "25mm", "25mm", "25mm" },
    { "A9L",		"52mm",  "37mm",   "25mm", "25mm", "25mm", "25mm" },
    { "A10L",		"37mm",  "26mm",   "25mm", "25mm", "25mm", "25mm" },

    /* ISO intermediate paper B-series (portrait) */

    { "B0",		"1000mm","1414mm", "25mm", "25mm", "25mm", "25mm" },
    { "B1",		"707mm", "1000mm", "25mm", "25mm", "25mm", "25mm" },
    { "B2",		"500mm", "707mm",  "25mm", "25mm", "25mm", "25mm" },
    { "B3",		"353mm", "500mm",  "25mm", "25mm", "25mm", "25mm" },
    { "B4",		"250mm", "353mm",  "25mm", "25mm", "25mm", "25mm" },
    { "B5",		"176mm", "250mm",  "25mm", "25mm", "25mm", "25mm" },
    { "B6",		"125mm", "176mm",  "25mm", "25mm", "25mm", "25mm" },

    /* ISO intermediate paper B-series (landscape) */

    { "B0L",		"1414mm","1000mm", "25mm", "25mm", "25mm", "25mm" },
    { "B1L",		"1000mm","707mm",  "25mm", "25mm", "25mm", "25mm" },
    { "B2L",		"707mm", "500mm",  "25mm", "25mm", "25mm", "25mm" },
    { "B3L",		"500mm", "353mm",  "25mm", "25mm", "25mm", "25mm" },
    { "B4L",		"353mm", "250mm",  "25mm", "25mm", "25mm", "25mm" },
    { "B5L",		"250mm", "176mm",  "25mm", "25mm", "25mm", "25mm" },
    { "B6L",		"176mm", "125mm",  "25mm", "25mm", "25mm", "25mm" },

    /* ISO envelopes for A-series */

    { "C0",		"1294mm","916mm",  "25mm", "25mm", "25mm", "25mm" },
    { "C1",		"916mm", "647mm",  "25mm", "25mm", "25mm", "25mm" },
    { "C2",		"647mm", "458mm",  "25mm", "25mm", "25mm", "25mm" },
    { "C3",		"458mm", "323mm",  "25mm", "25mm", "25mm", "25mm" },
    { "C4",		"323mm", "229mm",  "25mm", "25mm", "25mm", "25mm" },
    { "C5",		"229mm", "161mm",  "25mm", "25mm", "25mm", "25mm" },
    { "C6",		"161mm", "114mm",  "25mm", "25mm", "25mm", "25mm" },

    /* British paper sizes (some are also used in the USA) */

    { "Octavo",		"5in",   "8in",    "1in",  "1in",  "1in",  "1in" },
    { "Sixmo",          "6.5in", "8in",    "1in",  "1in",  "1in",  "1in" },
    { "Quarto",         "8in",   "10in",   "1in",  "1in",  "1in",  "1in" },
    { "Foolscap",	"8.5in", "13in",   "1in",  "1in",  "1in",  "1in" },
    { "Government-legal","8.5in","13in",   "1in",  "1in",  "1in",  "1in" },
    { "Folio",		"8.3in", "13in",   "1in",  "1in",  "1in",  "1in" },

    { CNULL,		CNULL,   CNULL,    CNULL,  CNULL,  CNULL,   CNULL}
};

#ifndef FASTIO
#define FASTIO		0
#endif /* FASTIO */

#if    OS_PCDOS
#include <io.h>
#endif /* OS_PCDOS */

#if    FASTIO
#include <jsys.h>
#include <mon_s.h>
#define FRead(buffer,itemcount,itemsize,fp) \
    ((size_t)_read(jfnof((int)fileno(fp),buffer,(int)(itemsize*itemcount))))
#define FWrite(buffer,itemcount,itemsize,fp) \
    ((size_t)_write(jfnof((int)fileno(fp),buffer,(int)(itemsize*itemcount))))
#else
#if ardent
					/* 3rd arg is int, sigh... */
#define FRead(buffer,itemcount,itemsize,fp) \
	((size_t)fread(buffer,itemcount,(int)itemsize,fp))
#define FWrite(buffer,itemcount,itemsize,fp) \
	((size_t)fwrite(buffer,itemcount,(int)itemsize,fp))
#else
#define FRead(buffer,itemcount,itemsize,fp) \
	((size_t)fread(buffer,itemcount,(size_t)itemsize,fp))
#define FWrite(buffer,itemcount,itemsize,fp) \
	((size_t)fwrite(buffer,itemcount,(size_t)itemsize,fp))
#endif
#endif /* FASTIO */

#if    (OS_TOPS20 || OS_UNIX || OS_VAXVMS)
#if    KCC_20
	/* KCC-20 uses magic cookies for ftell()/fseek() unless we
	have binary files; with the latter, we get normal byte offsets. */
#if 0
#define FOPEN_W "wb7C-"
#define FOPEN_R "rb7C-"
#define FOPEN_RW "w+b7C-"
#else  /* NOT 0 */
#define FOPEN_W "wb7"
#define FOPEN_R "rb7"
#define FOPEN_RW "w+b7"
#endif /* 0 */
#include <sys/usysio.h>	/* ugly hackery... */
#else  /* NOT KCC_20 */
#define FOPEN_W "w"
#define FOPEN_R "r"
#define FOPEN_RW "w+"
#endif /* KCC_20 */
#else  /* NOT (OS_TOPS20 || OS_UNIX || OS_VAXVMS) */
#define FOPEN_W "wb"
#define FOPEN_R "rb"
#define FOPEN_RW "w+b"
#endif /* (OS_TOPS20 || OS_UNIX || OS_VAXVMS) */


#ifdef FPRINTF
#undef FPRINTF
#endif
#define FPRINTF		(void)fprintf

#ifdef PRINTF
#undef PRINTF
#endif
#define	PRINTF		(void)printf

#ifdef SPRINTF
#undef SPRINTF
#endif
#define SPRINTF		(void)sprintf

#ifndef AFMENVVAR			/* can define at compile time */
#define AFMENVVAR	"AFMPATH"
#endif

#ifndef AFMEXT				/* can define at compile time */
#if OS_VMCMS
#define AFMEXT		" afm"
#else
#define AFMEXT		".afm"
#endif
#endif

#ifndef AFMMAPCOMMENTCHAR		/* can define at compile time */
#define AFMMAPCOMMENTCHAR	'%'
#endif

#ifndef AFMMAPFILE			/* can define at compile time */
#if OS_VMCMS
#define AFMMAPFILE	"afmfonts map"
#else
#define AFMMAPFILE	"afmfonts.map"
#endif
#endif

#ifndef AFMMAPSEPCHARS			/* can define at compile time */
#define AFMMAPSEPCHARS	" \t=:"		/* token separators in AFMMAPFILE */
#endif

#ifndef AFMPATH				/* can define at compile time */

#if OS_ATARI || OS_PCDOS
#define AFMPATH		"c:\\afm"
#endif

#if OS_PRIMOS
#define AFMPATH		"afm>"
#endif

#if OS_RMX
#define AFMPATH		"afm:"
#endif

#if OS_TOPS20
#define AFMPATH		"sys"
#endif

#if OS_UNIX
#define AFMPATH		"/usr/local/lib/afm"
#endif

#if OS_VAXVMS
#define AFMPATH		"sys$system"
#endif

#if OS_VMCMS
#define AFMPATH		""
#endif

#endif /* AFMPATH */

#define BOL()		((in_col == 0) && (in_str == 0) && (level == 0))
					/* test for beginning of line */

#ifndef FONTSEPCHARS			/* can define at compile time */
#define FONTSEPCHARS	" \t,"	/* font separators in -ffont1,font2,... */
#endif

#ifndef INITFILECOMMENTCHAR		/* can define at compile time */
#define INITFILECOMMENTCHAR	AFMMAPCOMMENTCHAR
#endif

#ifndef INITFILESEPCHARS		/* can define at compile time */

#endif

#ifndef INITENVVAR			/* can define at compile time */
#define INITENVVAR	AFMENVVAR	/* environment variable for INITFILE */
					/* search path */
#endif

#ifndef INITFILE			/* can define at compile time */
#if OS_VMCMS
#define INITFILE	"lptops ini"	/* default initfile name */
#else
#define INITFILE	"lptops.ini"
#endif
#endif

#ifndef INITPATH			/* can define at compile time */
#define INITPATH	AFMPATH		/* default initfile search path */
#endif

#define NUL		'\0'

/***********************************************************************
PostScript's basic  units are big points (72bp =  1in).   Everything  in
this routine works in terms of "Quite Small Units": 1 inch = QSU qsu.

For convenience in manual deciphering of the PostScript output, we  want
QSU to be a power of 10.

The unit should be large in order  to  avoid coordinate truncation error
accumulation  and  to  achieve  accurate  positioning  on  the  highest-
resolution typesetters (5300 lines per inch).   It must be small  enough
that character widths from font-metric files and page coordinates can be
computed without integer overflow.

With   32-bit  signed   integers,   the   largest   positive  number  is
2,147,483,647, and  assuming a  maximum page dimension  of 1414mm (55.67
inches)  (ISO  B0 forms),  the largest possible value for  QSU would  be
2,147,483,647 / 55.67 = 38,575,731.

Although  larger  values  of QSU  require  more coordinate digits in the
output PostScript, the impact on output  size is minimal,  because there
are only four numbers output for each page column.  All other coordinate
computations  are  hidden  in  macros.   Choosing  QSU  =  10**6  allows
dimensions up to  2147  inches (more  than  the  height  of  a  17-story
building),  and supports  a resolution  of  less than  the wavelength of
visible light.  TeX DVI file coordinates are in scaled points (65536sp =
1pt, 4,736,286.7sp = 1in), making 1 QSU = 4.736 DVI unit, so the choices
are similar.

***********************************************************************/

#define QSU 1000000L		/* Quite Small Units per inch (see above) */

#define QSU_TO_BP(x)	(72.0 * QSU_TO_IN(x))
				/* QSU to PostScript units (big points) */

#define QSU_TO_CM(x)	(2.54 * QSU_TO_IN(x)) /* QSU to cm */

#define QSU_TO_IN(x)	((double)(x) / (double)QSU) /* QSU to inches */

#define QSU_TO_PT(x)	(72.27 * QSU_TO_IN(x)) /* QSU to TeX points */

#define INCH_TO_QSU(x)	((COORDINATE)(0.5 + x * (double)QSU))
				/* inch to QSU conversion with rounding */

#define DIMEN_TO_QSU(s)	INCH_TO_QSU(inch(s))
				/* any dimension to QSU with rounding */

#define LEFT_X (left_margin + col_margin)
#define BOT_Y  (bottom_margin + col_margin)
#define TOP_Y(margin) (page_height - (margin) - baseline_skip - col_margin)

#define GETC(fp)	((npushback > 0) ? pushback[--npushback] : getc(fp))
#define UNGETCHAR(c)	((void)ungetchar((int)(c)))

#ifdef MAX
#undef MAX
#endif /* MAX */

#define MAX(a,b) ((a) > (b) ? (a) : (b))

#ifdef MIN
#undef MIN
#endif /* MIN */

#define MIN(a,b) ((a) < (b) ? (a) : (b))

#define MAXBLOCK 4096L		/* block size for file copy */
#define MAXLINEWIDTH 79		/* maximum output line width */
#define MAXPAGE 1024		/* size of page location table */
#define MAXPUSHBACK 5		/* how many characters of pushback */

#if    OS_TOPS20	/* want <CR><LF> */
#define NEWLINE(fp) {(void)putc((char)'\r',fp);(void)putc((char)'\n',fp);}
#define OUTNL()		{Out_Char('\r'); Out_Char('\n');}
#else  /* NOT OS_TOPS20 */
#define NEWLINE(fp) (void)putc((char)'\n',fp)	/* want bare <LF> */
#define OUTNL()		Out_Char('\n')	/* want bare <LF> */
#endif /* OS_TOPS20 */

#define OUTCHAR(c)	Out_Char((int)(c))

#define OUTSTR(s)	Out_String((const char*)(s))
#define OUTSTRCHAR(c)	Out_StrChar((int)(c))

#ifndef PAPERSEPCHARS		/* can define at compile time */
#define PAPERSEPCHARS	" \t,"	/* separators in -paper=name,width,height */
#endif

#ifndef OPTIONCHARS
#if (OS_TOPS20 || OS_VAXVMS)
#define OPTIONCHARS	"/-" /* command line options start with one of these */
#else
#define OPTIONCHARS	"-"
#endif
#endif

#define COL_TWIDDLE 40L		/* 1/fraction of column width for */
				/* column margin */

#define PAGE_T 001		/* flag bits for page_flag */
#define PAGE_B 002
#define PAGE_L 010
#define PAGE_R 020
#define PAGE_C 040

#if OS_VMCMS
#define PFAEXT	" pfa"
#define PFBEXT	" pfb"
#else
#define PFAEXT	".pfa"
#define PFBEXT	".pfb"
#endif

#define PFB_MAGIC	0x80
#define	PFB_TEXT	0x01
#define PFB_BINARY	0x02
#define PFB_EOF		0x03

static int		npushback = 0; /* for inchar(), ungetchar() */
static int		pushback[MAXPUSHBACK];
static char		tempfile_1[MAXPATHLEN+1] = "";
static char		tempfile_2[MAXPATHLEN+1] = "";

#if    PCC_20
/* PCC-20 has no tmpnam(), and KCC-20's tmpnam() always makes a file
in the current directory, which might not be writable. */
#define TEMPFILE(fn) (fn[0] ? fn : \
	(strcpy(fn,"ps:<temp>lptops.XXXXXX"),\
	(char*)mktemp(fn)))		/* temporary output file */
#endif /* PCC_20 */

#if    KCC_20
#define TEMPFILE(fn) (fn[0] ? fn : \
	(strcpy(fn,"ps:<temp>"),\
	strcat(fn,tmpnam((char*)NULL)))) /* temporary output file */
#endif /* KCC_20 */

#ifndef TEMPFILE
#if defined(HAVE_MKSTEMP) && defined(HAVE_MKSTEMP_DECL) && defined(OS_UNIX)
#define TEMPFILE(fn) fn
#else
#define TEMPFILE(fn) (fn[0] ? fn : \
	(strcpy(fn,(char*)tmpnam((char*)NULL))))
					/* temporary output file */
#endif /* defined(HAVE_MKSTEMP) && defined(HAVE_MKSTEMP_DECL) && defined(OS_UNIX) */
#endif /* TEMPFILE */

#define TEMPFILE_1	TEMPFILE(tempfile_1)
#define TEMPFILE_2	TEMPFILE(tempfile_2)

#define COMPARE_CAST	int (*) ARGS((const char*, const char*))

				/* global functions */

static void		Beg_Column ARGS((void));
static void		Beg_Line ARGS((void));
static void		Beg_Page ARGS((void));
static void		Beg_String ARGS((void));
static void		Beg_Text ARGS((void));
static void		Check_Page_Selection ARGS((int n__));
static void		Copy_Block ARGS((FILE *fpin_, FILE *fpout_,
			    long b1_, long b2_));
static void		Copy_Plist ARGS((FILE *fp_));
static void		Copy_PFA ARGS((FILE *fpout_, FILE *fpin_));
static void		Copy_PFB ARGS((FILE *fpout_, FILE *fpin_));
static void		Copy_PFB_Binary ARGS((FILE *fpout_, FILE *fpin_,
			    size_t block_length_));
static void		Copy_PFB_Text ARGS((FILE *fpout_, FILE *fpin_,
			    size_t block_length_));
static void		Do_All_Files ARGS((int argc_, char *argv_[]));
static void		Do_Comments ARGS((int argc_, char *argv_[]));
static void		Do_File ARGS((FILE *fp_));
static void		Do_Fonts ARGS((void));
static void		Do_Initialization ARGS((void));
static void		Do_ISO8859_1_encoding ARGS((void));
static void		Do_Options ARGS((int argc_, char *argv_[]));
static void		Do_PageSetup ARGS((void));
static void		Do_PageTable ARGS((void));
static void		Do_Prolog ARGS((void));
static void		Do_Resources ARGS((void));
static void		Do_Scribe_string ARGS((FILE *fp_, int c_));
static void		Do_Separator ARGS((void));
static void		Do_Setup ARGS((void));
static void		Do_Startup_File ARGS((void));
static void		Do_Tempfile ARGS((void));
static void		Do_Trailer ARGS((void));
static void		Do_Validation ARGS((void));
static BOOLEAN		Download_Font ARGS((const char *font_name));
static void		Dump_Font ARGS((const char* ps_name_,
			    const char* ps_fonts_[]));
static void		Dump_Paper ARGS((void));
static void		End_Column ARGS((void));
static void		End_Page ARGS((void));
static void		End_Line ARGS((void));
static void		End_String ARGS((void));
static void		End_Text ARGS((void));
static void		Fatal ARGS((const char *fmt_, const char *arg_));
static FILE		*Fopen_Tempfile ARGS((char * filename_, const char *mode_));
extern char 		*findfile ARGS((const char *pathlist_,
			    const char *name_));
static int		Get_Bold_Char ARGS((FILE * fp_));
static int		Get_Underlined_Char ARGS((FILE * fp_));
static char		*Get_Wrapped_Line ARGS((char *line_, size_t maxline_,
			    FILE *fpin_));

#if    OS_UNIX
char			*getupn ARGS((void));
#endif /* OS_UNIX */

extern double		inch ARGS((const char *s_));
static int		inchar ARGS((FILE *fp_));
int			main ARGS((int argc_, char *argv_[]));
static char		*Map_Font ARGS((const char *font_name_,
			    const char *ext_));
static void		Option ARGS((const char *opt_));
static void		Opt_Author ARGS((const char* option_,
			    const char *value_));
static void		Opt_Copyright ARGS((const char* option_,
			    const char *value_));
static void		Opt_Error ARGS((const char *fmt_, const char *option_));
static void		Opt_Font ARGS((const char* option_,const char *value_));
static void		Opt_Help ARGS((const char* option_,const char *value_));
static void		Opt_Page ARGS((const char* option_,const char *value_));
static void		Opt_Paper ARGS((const char* option_,
			    const char *value_));
static void		Opt_SelectPages ARGS((const char* option_,
			    const char *value_));
static void		Opt_Version ARGS((const char* option_,
			    const char *value_));
static void		Out_Atsign ARGS((FILE *fp_));
static void		Out_Backspace ARGS((void));
static void		Out_Bullet ARGS((void));
static void		Out_Char ARGS((int c_));
static void		Out_Escape ARGS((FILE *fp_));
static void		Out_Fontdef ARGS((const char *macro_,
			    const char *font_));
static void		Out_Font_Resource ARGS((const char *name_,
			    int nfonts_));
static void		Out_Formfeed ARGS((void));
static void		Out_Lines ARGS((FILE *fp_, const char *lines_[]));
static int		Out_Manchar ARGS((FILE *fp_, int c_));
static void		Out_Need ARGS((int count_));
static void		Out_Newline ARGS((void));
static void		Out_Other ARGS((int c_));
static void		Out_Return ARGS((FILE *fp_));
static void		Out_StrChar ARGS((int c_));
static void		Out_String ARGS((const char *s_));
static void		Out_Tab ARGS((void));
static void		Out_Underline ARGS((FILE *fp_));
static void		Out_Vskip ARGS((double v_));
static size_t		PFB_Length ARGS((FILE *fpin));
static void		Read_Metrics ARGS((Font *font_));
static void		Reencode ARGS((void));
static void		Reverse_Pages ARGS((void));
static void		Set_Family ARGS((const char *name_normal,
			    BOOLEAN resident_normal,
			    const char *name_bold,
			    BOOLEAN resident_bold,
			    const char *name_italic,
			    BOOLEAN resident_italic,
			    const char *name_both,
			    BOOLEAN resident_both,
			    const char *name_typewriter,
			    BOOLEAN resident_typewriter));
static void		Strip_Comments ARGS((char line_[], int comment_char_));
static int		Tokenize ARGS((const char *s_, const char *name_[],
			    int maxname_, const char *sepchars_));
extern int		tscanf ARGS((const char *str_,const char *fmt_,...));
static int		ungetchar ARGS((int c_));
static void		Usage ARGS((int exit_status));
static void		Warning ARGS((const char *fmt_, const char *arg_));
static void		Wrap_Check ARGS((int c_));

				/* global variables */


static const char *	afmpath;	/* font search path */

static BOOLEAN 		bad_option;	/* bad option flag */
static COORDINATE	baseline_skip;	/* inter-line spacing */
static Font		bold;		/* bold font structure */
static Font		bolditalic;	/* bold-italic font structure */
static COORDINATE	bottom_margin;	/* bottom margin width in QSU */

static BOOLEAN		clip;		/* TRUE for column-clipping */
static int		column;		/* current text column (1..numcol) */
static COORDINATE	col_margin;	/* column margin (extra space */
					/* between column borders in */
					/* multi-column mode) */
static COORDINATE	col_width;	/* column width in QSU */
static BOOLEAN		comment_line;	/* current output line is a comment */
static int		copies;		/* number of copies of each page */
static Font *		current_font;	/* pointer to current font structure */
static int		current_id;	/* current font letter ('B', 'I', 'N',
					   'P', 'T', 'U') */

static BOOLEAN		do_page_number;	/* page numbering flag */

static int		in_col;		/* input line column */
static int		in_str;		/* current (...) string length */
static int		init_page_number; /* initial page number */
static COORDINATE	init_top_margin; /* initial page top margin in QSU */
static BOOLEAN		ISO8859_1_charset; /* ISO8859-1 character set, or
					Adobe Standard PostScript set? */
static Font		italic;		/* italic font structure */

static BOOLEAN		landscape_mode;	/* horizontal landscape mode output */
static COORDINATE	left_margin;	/* left margin in QSU */
static int		level;		/* parenthesis level */
static int		line_number;	/* current line number */
static COORDINATE	line_width;	/* current line width in QSU */

static BOOLEAN		manpage;	/* UNIX manual page input */
static int		manual_feed;	/* manual feed page count */
static const char *	msg_prefix;	/* error and warning message prefix */

static int		npages;		/* number of entries in pages[] */
static Font		normal;		/* normal (roman) font structure */
static int		numcol;		/* number of text columns */

static COORDINATE	outline;	/* page and column outline linewidth */
static int		out_column;	/* output column */
static BOOLEAN		output_this_page = TRUE;
					/* maybe FALSE if -page:n-n specified */

static int		page_count;	/* number of pages actually output */
static int		page_flag;	/* page number location flag */
static COORDINATE	page_height;	/* page height in QSU */
static int		page_number;	/* current page number */
static PAGE_SELECTION	*pages;/* output page selections */
static long		page_table[MAXPAGE+1];	/* page location table (with */
					/* space for extra entry marking EOF) */
static COORDINATE	page_width;	/* page width in QSU */
static const char *	paper_name;	/* name of paper form */
static COORDINATE	point_size;	/* font point size in big points */
static const char *	program_name;	/* global copy of argv[0] */

static BOOLEAN		quiet;		/* quiet flag */

static BOOLEAN		reverse;	/* page reversal flag */
static COORDINATE	right_margin;	/* right margin in QSU */

static BOOLEAN		scribe;		/* recognize bold/italic sequences */

static char		tempstr[80];	/* temporary for formatted output */
static FILE *		tf;		/* temporary output file */
static COORDINATE	top_margin;	/* top margin in QSU */
static Font		typewriter;	/* typewriter font structure */

static BOOLEAN		wrap;		/* TRUE->long lines wrap to new line */

static COORDINATE	x;		/* current horizontal coordinate */
static COORDINATE	x_right;	/* x coord. of right margin in QSU */

static COORDINATE	y;		/* current vertical coordinate */

/***********************************************************************
Built-in PostScript font tables.  The first column in each list contains
the full (case sensitive) PostScript  names, the second column  contains
convenient short abbreviations  which can  be used in  the command  line
-ffontname option,  the third  and fourth  columns contain  ditroff  and
Adobe TranScript abbreviations.

The  first  four  lines  of   entries  are  available  on  the   regular
LaserWriter, and the  remainder are  available only  on the  LaserWriter
Plus.
***********************************************************************/

#ifndef MAXEXTRAFONTS
#define MAXEXTRAFONTS	100		/* limit on user-definable families */
#endif

#define MAXFONTS	(4*(13 + MAXEXTRAFONTS))
					/* standard ones, and run-time extras */

#define PS_DEFAULT 0	/* index of default font in ps_normal[] */

#define PS_RESIDENT (12*4)		/* index of last resident font in */
					/* ps_normal[] */

static const char  *ps_normal[MAXFONTS] =
{
    /* fullname				short	ditroff	TranScript */
    "Courier",				"C",	"C",	"C",
    "Helvetica",			"H",	"he",	"he",
    "Times-Roman",			"T",	"R",	"ti",
    "Symbol",				"S",	"S",	"S",
    "AvantGarde-Book",			"AGB",	"ag",	"ag",
    "AvantGarde-Demi",			"AGD",	"Ag",	"Ag",
    "Bookman-Demi",			"BD",	"Bo",	"Bo",
    "Bookman-Light",			"BL",	"bo",	"bo",
    "Helvetica-Narrow",			"HN",	"hn",	"hn",
    "NewCenturySchlbk-Roman",		"NCSR",	"nc",	"nc",
    "Palatino-Roman",			"PR",	"pa",	"pa",
    "ZapfChancery-MediumItalic",	"ZCMI",	"ZC",	"ZC",
    "ZapfDingbats",			"ZD",	"ZD",	"ZD",
    CNULL,				CNULL,  CNULL,  CNULL
};

static const char  *ps_bold[MAXFONTS] =
{
    /* fullname				short	ditroff	TranScript */
    "Courier-Bold",			"CB",	"CB",	"CB",
    "Helvetica-Bold",			"HB",	"He",	"He",
    "Times-Bold",			"TB",	"B",	"Ti",
    "Symbol",				"S",	"S",	"S",
    "AvantGarde-Book",			"AGB",	"ag",	"ag",
    "AvantGarde-Demi",			"AGD",	"Ag",	"Ag",
    "Bookman-Demi",			"BD",	"Bo",	"Bo",
    "Bookman-Light",			"BL",	"bo",	"bo",
    "Helvetica-Narrow-Bold",		"HNB",	"Hn",	"Hn",
    "NewCenturySchlbk-Bold",		"NCSB",	"Nc",	"Nc",
    "Palatino-Bold",			"PB",	"Pa",	"Pa",
    "ZapfChancery-MediumItalic",	"ZCMI",	"ZC",	"ZC",
    "ZapfDingbats",			"ZD",	"ZD",	"ZD",
    CNULL,				CNULL,  CNULL,  CNULL
};

static const char  *ps_italic[MAXFONTS] =
{
    /* fullname				short	ditroff	TranScript */
    "Courier-Oblique",			"CO",	"CO",	"CO",
    "Helvetica-Oblique",		"HO",	"hE",	"hE",
    "Times-Italic",			"TI",	"I",	"tI",
    "Symbol",				"S",	"S",	"S",
    "AvantGarde-BookOblique",		"AGBO",	"aG",	"aG",
    "AvantGarde-DemiOblique",		"AGDO",	"AG",	"AG",
    "Bookman-DemiItalic",		"BDI",	"BO",	"BO",
    "Bookman-LightItalic",		"BLI",	"bO",	"bO",
    "Helvetica-Narrow-Oblique",		"HNO",	"hN",	"hN",
    "NewCenturySchlbk-Italic",		"NCSI",	"nC",	"nC",
    "Palatino-Italic",			"PI",	"pA",	"pA",
    "ZapfChancery-MediumItalic",	"ZCMI",	"ZC",	"ZC",
    "ZapfDingbats",			"ZD",	"ZD",	"ZD",
    CNULL,				CNULL,  CNULL,  CNULL
};

static const char  *ps_both[MAXFONTS] =
{
    /* fullname				short	ditroff	TranScript */
    "Courier-BoldOblique",		"CBO",	"CD",	"CD",
    "Helvetica-BoldOblique",		"HBO",	"HE",	"HE",
    "Times-BoldItalic",			"TBI",	"BI",	"TI",
    "Symbol",				"S",	"S",	"S",
    "AvantGarde-BookOblique",		"AGBO",	"aG",	"aG",
    "AvantGarde-DemiOblique",		"AGDO",	"AG",	"AG",
    "Bookman-DemiItalic",		"BDI",	"BO",	"BO",
    "Bookman-LightItalic",		"BLI",	"bO",	"bO",
    "Helvetica-Narrow-BoldOblique",	"HNBO",	"HN",	"HN",
    "NewCenturySchlbk-BoldItalic",	"NCSBI","NC",	"NC",
    "Palatino-BoldItalic",		"PBI",	"PA",	"PA",
    "ZapfChancery-MediumItalic",	"ZCMI",	"ZC",	"ZC",
    "ZapfDingbats",			"ZD",	"ZD",	"ZD",
    CNULL,				CNULL,  CNULL,  CNULL
};

static const char  *ps_typewriter[MAXFONTS] =
{
    /* fullname				short	ditroff	TranScript */
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    "Courier",				"C",	"C",	"C",
    CNULL,				CNULL,  CNULL,  CNULL
};

#define T_INT		1
#define T_BOOLEAN	2
#define T_DIMEN		3
#define T_SPECIAL	4

typedef struct
{
    const char	*name;
    size_t	min_match;
    int		type;
    const char	*default_value;
    void	*variable;
#if 0
    /* This is Standard C */
    void	(*special) ARGS((const char *option_,const char *s_));
#else
    /* This is for brain-damaged sub-Standard C compilers that cannot
       handle variable names in function prototypes: DEC Ultrix 4.3,
       and possibly others, sigh... */
    void	(*special) ARGS((const char *,const char *));
#endif
} OPTION;

static OPTION Option_Table[] =
{
    { "?",		1, T_SPECIAL,	CNULL,	VNULL,		Opt_Help},
    { "author",  	1, T_SPECIAL,	CNULL,	VNULL,		Opt_Author},
    { "bottom",	  	1, T_DIMEN,	"1in",	&bottom_margin,	FNULL},
    { "c",	  	1, T_INT,	"1",	&copies,	FNULL},
    { "clip",	  	2, T_BOOLEAN,	CNULL,	&clip,		FNULL},
    { "copyleft",	5, T_SPECIAL,	CNULL,	VNULL,		Opt_Copyright},
    { "copyright",	5, T_SPECIAL,	CNULL,	VNULL,		Opt_Copyright},
    { "e",		1, T_BOOLEAN,	CNULL,	&ISO8859_1_charset, FNULL},
    { "font",		1, T_SPECIAL,	"Courier", VNULL,	Opt_Font},
    { "help",		1, T_SPECIAL,	CNULL,	VNULL,		Opt_Help},
    { "i",		1, T_DIMEN,	"1in",	&init_top_margin,FNULL},
    { "landscape",	2, T_BOOLEAN,	CNULL,	&landscape_mode,FNULL},
    { "left",		1, T_DIMEN,	"1in",	&left_margin,	FNULL},
    { "m",		1, T_INT,	"1",	&numcol,	FNULL},
    { "manpage",	2, T_BOOLEAN,	CNULL,	&manpage,	FNULL},
    { "n",		1, T_SPECIAL,	CNULL,	VNULL,		Opt_Page},
    { "o",		1, T_DIMEN,	"0.4pt",&outline,	FNULL},
    { "p",		1, T_DIMEN,	"10bp",	&point_size,	FNULL},
    { "pages",		3, T_SPECIAL,	"",	VNULL, Opt_SelectPages},
    { "paper",		3, T_SPECIAL,	"A",	VNULL,		Opt_Paper},
    { "quiet",		1, T_BOOLEAN,	CNULL,	&quiet,		FNULL},
    { "reverse",	2, T_BOOLEAN,	CNULL,	&reverse,	FNULL},
    { "right",		1, T_DIMEN,	"1in",	&right_margin,	FNULL},
    { "s",		1, T_INT,	"32767", &manual_feed,	FNULL},
    { "scribe",		2, T_BOOLEAN,	CNULL,	&scribe,	FNULL},
    { "top",		1, T_DIMEN,	"1in",	&top_margin,	FNULL},
    { "version",	1, T_SPECIAL,	CNULL,	VNULL,		Opt_Version},
    { "wrap",		1, T_BOOLEAN,	CNULL,	&wrap,		FNULL},
    { "y",		1, T_DIMEN,	"0bp",	&baseline_skip,	FNULL},
    { CNULL,		0, 0,		CNULL,	VNULL,		FNULL}
};

int
#if STDC
main(
int				argc,
char				*argv[]
)
#else /* NOT STDC */
main(argc,argv)
int				argc;
char				*argv[];
#endif /* STDC */
{
#if    OS_VAXVMS
    /* Provide a UNIX-like command-line interface with wildcard */
    /* expansion and input and output file redirection.*/

    extern char			**cmd_lin();

    argv = cmd_lin( "", &argc, argv );
#endif /* OS_VAXVMS */

    program_name = (argv[0] != CNULL) ? argv[0] : "lptops";

    Do_Initialization();
    Do_Options(argc,argv);
    Do_Validation();
    Do_Tempfile();
    Do_Comments(argc,argv);
    Do_Prolog();
    Do_Setup();
    Do_All_Files(argc,argv);
    if (reverse)
	Reverse_Pages();
    Do_Trailer();

    if (pages != (PAGE_SELECTION*)NULL)
	FREE(pages);

    exit(EXIT_SUCCESS);
    return (0);				/* keep lint and compilers happy */
}

static void
Beg_Column(VOID_ARG)
{
    in_col = 0;
    in_str = 0;
    level = 0;
    line_number = 0;
    line_width = 0L;

    OUTSTR("CB");

#if    FIX31OCT85
    OUTSTR("()pop % no-op for PostScript XON/XOFF bug workaround");
#endif /* FIX31OCT85 */

    OUTNL();


    SPRINTF(tempstr,"%ld %ld NEWCOLUMN",
	(long)(x - col_margin),(long)bottom_margin);
    OUTSTR(tempstr);
    OUTNL();

    SPRINTF(tempstr,"%ld %ld moveto",x,y);
    OUTSTR(tempstr);
    OUTNL();

    if (!(manpage || scribe))
    {					/* select normal font */
	OUTSTR("FN");
	OUTNL();
    }
}

static void
Beg_Line(VOID_ARG)
{
    if (manpage || scribe)
	OUTCHAR('S');		/* gsave matched by grestore from End_Line() */
    in_col = 0;
    in_str = 0;
    level = 0;
}

static void
Beg_Page(VOID_ARG)
{
    int				k;

    column = 1;
    (void)fflush(tf);
    k = page_number-init_page_number;
    if (k < MAXPAGE)
	page_table[k] = (long)ftell(tf);
    Check_Page_Selection(page_number);
    SPRINTF(tempstr,"%%%%Page: %d %d",k+1,page_number);
    OUTSTR(tempstr);
    OUTNL();
    Do_PageSetup();
    if (!quiet && output_this_page)
	FPRINTF(stderr,
	    " [%d",page_number);/* begin [p.l.l...] status report */
}

static void
Beg_String(VOID_ARG)
{
    if (BOL())
	Beg_Line();
    End_Text();
    OUTCHAR('(');
    level++;
    in_str = 0;
}

static void
Beg_Text(VOID_ARG)
{
    Beg_String();
}

#if STDC
static void
Check_Page_Selection(int n)
#else /* NOT STDC */
static void
Check_Page_Selection(n)
int 				n;
#endif /* STDC */
{
    register int		k;	/* loop index */

    if (npages > 0)
    {
	output_this_page = FALSE;
	for (k = 0; k < npages; ++k)
	{
	    if ((pages[k].begin <= n) &&
		(n <= pages[k].end) &&
		(((n - pages[k].begin) % pages[k].step) == 0))
	    {
		output_this_page = TRUE;
		break;
	    }
	}
    }
    else
	output_this_page = TRUE;
}

static void
#if STDC
Copy_Block(
FILE				*fpin,	/* input file */
FILE				*fpout,	/* output file */
long				b1,	/* start,end byte positions */
long				b2
)
#else /* NOT STDC */
Copy_Block(fpin,fpout,b1,b2)	/* copy bytes b1..b2 from fpin to fpout */
FILE				*fpin;	/* input file */
FILE				*fpout;	/* output file */
long				b1;	/* start,end byte positions */
long				b2;
#endif /* STDC */
{
    static char			block[MAXBLOCK];
				/* static to save stack space on PC DOS */
    long			k;	/* loop index */
    size_t			length;	/* number of bytes to copy */

#if    FASTIO
    (void)fflush(fpout);		/* empty any buffered output */
#endif /* FASTIO */

    for (k = b1; k <= b2; k += MAXBLOCK)
    {
	length = (size_t)MIN(MAXBLOCK,b2-k+1);

	(void)fseek(fpin,k,0);		/* position to desired block */

	if (FRead(block,1,length,fpin) != length)
	    Fatal("Input block length error",CNULL);

	if (FWrite(block,1,(size_t)length,fpout) != length)
	    Fatal("Output block length error",CNULL);
    }
    (void)fflush(fpout);
}

static void
#if STDC
Copy_Plist(
FILE				*fp
)
#else /* NOT STDC */
Copy_Plist(fp)			/* copy parenthesized list to tf */
FILE				*fp;
#endif /* STDC */
{
    register int		c;
    register int		match;
    static char			p_open[]  = {'(','[','<','{','\'','\"','`',NUL};
    static char			p_close[] = {')',']','>','}','\'','\"','`',NUL};

    c = inchar(fp);
    match = 0;
    while (p_open[match] && (p_open[match] != (char)c))
	match++;

    if (!p_open[match])
	Fatal("@<letter> not followed by open parenthesis!", CNULL);

    Beg_Text();
    while (((c = inchar(fp)) != EOF) && ((char)c != p_close[match]))
    {
	if ((c < ' ') && (c != '\t'))
				/* avoid iscntrl() because implementations */
				/* vary in their treatment of 0200..0237 */
	    Fatal("@<letter>[...] contains control character!", CNULL);
	if (c == '@')
	{
	    c = inchar(fp);		/* look ahead one character */
	    if (c != (int)'@')
	    {
		UNGETCHAR(c);
		c = (int)'@';
	    }
	}
	Out_Other(c);
    }				/* end while */
    End_Text();
}

static void
#if STDC
Copy_PFA(
FILE				*fpout,
FILE				*fpin
)
#else /* NOT STDC */
Copy_PFA(fpout,fpin)
FILE				*fpout;
FILE				*fpin;
#endif /* STDC */
{
    char			buffer[BUFSIZ];
    size_t			nleft;
    size_t			nread;
    size_t			ntotal;
    size_t			nwrite;
    char			*p;

    ntotal = 0;
    while ((nread = FRead(buffer, sizeof(char), sizeof(buffer), fpin)) > 0)
    {	/* some systems do not write requested amount, so for safety,
	   we need a loop here. */
	ntotal += nread;
	for (p = buffer, nleft = nread; p < (buffer + nread);
	    p += nwrite, nleft -= nwrite)
	    nwrite = FWrite(p, sizeof(char), nleft, fpout);
    }
}

/***********************************************************************
.pfb (PostScript Font -- Binary) files contain a mixture of text and
binary data.  Generally, the PostScript prolog and comments are in text,
and the encrypted PostScript character descriptions are in binary.

.pfa (PostScript Font -- ASCII) files are equivalent, except that all
data are in text form, with the binary data expressed in hexadecimal.

.pfa files are about twice as big as their corresponding .pfb files.
Here is a table for the Bitstream Courier and Charter fonts:

=================================================================
Filename        Bytes   Filename        Bytes   Ratio (.pfa/.pfb)
=================================================================
bchb.pfa        66525   bchb.pfb        33799       1.968
bchbi.pfa       69409   bchbi.pfb       35229       1.970
bchr.pfa        68700   bchr.pfb        34869       1.970
bchri.pfa       69191   bchri.pfb       35118       1.970
bcrb.pfa        79136   bcrb.pfb        40008       1.978
bcrbi.pfa       78837   bcrbi.pfb       39871       1.977
bcrr.pfa        80678   bcrr.pfb        40766       1.979
bcrri.pfa       78113   bcrri.pfb       39511       1.977
=================================================================

However, binary files are less portable, and cause problems with some
spoolers.  Also, .pfb files contain additional record type and record
length fields that prevent their direct incorporation in PostScript
output by simple file copying.

For portability, we convert a .pfb file here to text form, with
hexadecimal lines of 64 characters each for the binary sections.  The
results should be identical to what would have been obtained by
inclusion of the corresponding .pfa file.

A .pfb section consists of a two-byte header, a 4-byte length field, and
`length' bytes of text or binary data.  In the case of text sections,
the line separators may be CR or LF or CR LF; we translate them all to
whatever the NEWLINE() macro produces.

The two-byte header consists of 0x80 0x0[123] (0x01 == text, 0x02 ==
binary, 0x03 == end-of-file).  The length field contains the
least-significant byte first.
***********************************************************************/

static void
#if STDC
Copy_PFB(
FILE				*fpout,
FILE				*fpin
)
#else /* NOT STDC */
Copy_PFB(fpout,fpin)
FILE				*fpout;
FILE				*fpin;
#endif /* STDC */
{
    int				c;

    for (;;)
    {
	c = fgetc(fpin);
	if (c == PFB_MAGIC)		/* then get block type code */
	    c = fgetc(fpin);
	switch (c)			/* copy one block of specified type */
	{
	case PFB_BINARY:
	    Copy_PFB_Binary(fpout,fpin,PFB_Length(fpin));
	    break;
	case PFB_EOF:
	    return;
	case PFB_TEXT:
	    Copy_PFB_Text(fpout,fpin,PFB_Length(fpin));
	    break;
	default:
	    Warning("Invalid .pfb font file","");
	    return;
	}
    }
}

static void
#if STDC
Copy_PFB_Binary(
FILE				*fpout,
FILE				*fpin,
size_t				block_length
)
#else /* NOT STDC */
Copy_PFB_Binary(fpout, fpin, block_length)
FILE				*fpout;
FILE				*fpin;
size_t				block_length;
#endif /* STDC */
{
    int				c;	/* current input character */
    int				col;	/* current line column */
    static char			hexchars[] = "0123456789ABCDEF";

    for (col = 0; block_length > 0; block_length--, col += 2)
    {
	c = fgetc(fpin);
	if (col >= 64)		/* wrap lines every 64 columns */
	{
	    NEWLINE(fpout);
	    col = 0;
	}
	(void)fputc(hexchars[0x0f & (c >> 4)], fpout);
	(void)fputc(hexchars[0x0f & c], fpout);
    }
    NEWLINE(fpout);
}

static void
#if STDC
Copy_PFB_Text(
FILE				*fpout,
FILE				*fpin,
size_t				block_length
)
#else /* NOT STDC */
Copy_PFB_Text(fpout, fpin, block_length)
FILE				*fpout;
FILE				*fpin;
size_t				block_length;
#endif /* STDC */
{
    int				c;

    for ( ; block_length > 0; block_length--)
    {
	if ((c = fgetc(fpin)) == '\r')
	{
	    c = ungetc(fgetc(fpin),fpin); /* peek ahead one character */
	    if (c == '\n')
		/* NO-OP */;		/* discard CR of CR LF pair */
	    else
		NEWLINE(fpout);		/* translate CR to LF */
	}
	else if (c == EOF)
	    return;			/* should never happen */
	else
	    (void)fputc(c,fpout);
    }
}

static void
#if STDC
Do_All_Files(
int				argc,
char				*argv[]
)
#else /* NOT STDC */
Do_All_Files(argc,argv)
int				argc;
char				*argv[];
#endif /* STDC */
{
    FILE			*fp;	/* input file pointer */
    register int		k;	/* loop index */
    register int		nfiles;	/* count of file arguments */

    current_id = (manpage || scribe) ? (int)'N' : (int)'P';

    nfiles = 0;				/* count of file names */
    for (k = 1; k < argc; ++k)
    {
	if (strcmp(argv[k],"-") == 0)	/* magic value for stdin */
	{
	    nfiles++;
	    Do_File(stdin);
	    errno = 0;			/* clear global error flag */
	}
	else if (strchr(OPTIONCHARS,argv[k][0]) == (char *)NULL)
	{				/* not option, must be file name */
	    nfiles++;			/* count file name */
	    if ((fp = fopen(argv[k],FOPEN_R)) != (FILE *)NULL)
	    {
		Do_File(fp);
		(void)fclose(fp);
	    }
	    else
		Warning("Open failure on [%s]---file skipped", argv[k]);
	    errno = 0;			/* clear global error flag */
	}
    }
    if (nfiles == 0)
	Do_File(stdin);
}

static void
#if STDC
Do_Comments(
int				argc,
char				*argv[]
)
#else /* NOT STDC */
Do_Comments(argc,argv)
int				argc;
char				*argv[];
#endif /* STDC */
{
    static char			empty_string[] = "";

#if OS_UNIX
    char			hostname[256];
#endif /* OS_UNIX */

    register size_t		k;	/* loop index */
    char			timestr[27];	/* result of ctime() */
    long			the_timeval;	/* result of time() */
    char			*username;

    /*******************************************************************
    Make the PostScript File Structure header according to Version  1.0,
    First Edition  (January 1985)  of  the PostScript  File  Structuring
    Conventions manual.  We add  one additional field  at the end  after
    the %%Pages: n field:

    %%PageTable: m1 n1 p1 ... mk nk pk

    where the triples (mj  nj pj) contain the  page number pair (mj  nj)
    like the field

    %%Page: mj nj

    and pj is the  byte position in  the file of  the beginning of  that
    "%%Page" command.  A spooler can  then easily position to  arbitrary
    pages in the file for reprinting.
    *******************************************************************/

    OUTSTR("%!PS-Adobe-3.0");		/* magical file header */
    OUTNL();

    OUTSTR("%%Title:");			/* this contains our command line */
    for (k = 0; k < (size_t)argc; ++k)
    {
	if ((out_column + 3 + (int)strlen(argv[k])) > MAXLINEWIDTH)
	{				/* wrap long title comment */
	    OUTNL();
	    OUTSTR("%%+\t");
	}
	OUTCHAR(' ');
	if (k > 0)
	    OUTCHAR('\'');
	OUTSTR(argv[k]);
	if (k > 0)
	    OUTCHAR('\'');
    }
    OUTNL();			/* end of %%Title line */

    the_timeval = time((time_t *)NULL);
    (void)strcpy(timestr,ctime((CTIME_CAST)&the_timeval));
    k = strlen(timestr) - 1;	/* ctime has its own \n */
    timestr[k] = NUL;		/* so kill it */
    OUTSTR("%%CreationDate: ");
    OUTSTR(timestr);
    OUTNL();

    OUTSTR("%%Creator: ");
    OUTSTR(VERSION);
    OUTNL();

    OUTSTR("%%+\tcompiled ");

#if defined(__DATE__) && defined(__TIME__)
    OUTSTR("on ");
    OUTSTR(__DATE__);
    OUTSTR(" ");
    OUTSTR(__TIME__);
#endif /* defined(__DATE__) && defined(__TIME__) */

#if defined(USER)
    OUTSTR(" by ");
    OUTSTR(USER);
#endif

#if defined(HOSTNAME)
    OUTSTR("@");
    OUTSTR(HOSTNAME);
#endif

    OUTNL();				/* end "%%+\tcompiled" */

    /* Under TOPS-20, JSYS GFUST% cannot be used to obtain file owner string
    of a terminal file, which stdout always is, so just use job login name */

    OUTSTR("%%For: ");
    username = cuserid((char*)NULL);
    username = (username == (char*)NULL) ? empty_string : username;

#if OS_UNIX
    OUTSTR(getupn());
    OUTSTR(" <");
    OUTSTR(username);
    if (gethostname(hostname,sizeof(hostname)) == 0)
    {
	hostname[sizeof(hostname)-1] = NUL;/* BSD doesn't ensure final NULL */
	if (hostname[0])
	{
	    OUTSTR("@");
	    OUTSTR(hostname);
	}
    }
    OUTSTR(">");
#else /* NOT OS_UNIX */
    OUTSTR(username);
#endif /* OS_UNIX */

    OUTNL();

    /* If clipping is in effect, then we can guarantee the bounding box */
    /* extents.  Otherwise, we just assume full-page boundaries. */

    if (clip)
	FPRINTF(tf,"%%%%BoundingBox: %d %d %d %d",
	       (int)floor(QSU_TO_BP(left_margin)),
	       (int)floor(QSU_TO_BP(bottom_margin)),
	       (int)ceil(QSU_TO_BP(page_width - right_margin)),
	       (int)ceil(QSU_TO_BP(page_height - top_margin)));
    else
	FPRINTF(tf,"%%%%BoundingBox: %d %d %d %d",
	       (int)floor(QSU_TO_BP(0L)),
	       (int)floor(QSU_TO_BP(0L)),
	       (int)ceil(QSU_TO_BP(page_width)),
	       (int)ceil(QSU_TO_BP(page_height)));
    OUTNL();

    FPRINTF(tf,"%%%%DocumentMedia: %s %d %d 0 () ()",
	paper_name,
	(int)ceil(QSU_TO_BP(page_width)),
	(int)ceil(QSU_TO_BP(page_height)));
    OUTNL();

    OUTSTR("%%DocumentData: Clean7Bit");
    OUTNL();

    OUTSTR("%%DocumentFonts: font ");
    OUTSTR(normal.name);
    OUTNL();

    if (manpage || scribe)
    {
	OUTSTR("%%+\tfont ");
	OUTSTR(bold.name);
	OUTNL();

	OUTSTR("%%+\tfont ");
	OUTSTR(italic.name);
	OUTNL();
    }

    if (scribe)
    {
	OUTSTR("%%+\tfont ");
	OUTSTR(bolditalic.name);
	OUTNL();

	OUTSTR("%%+\tfont ");
	OUTSTR(typewriter.name);
	OUTNL();
    }

    if (manpage)
    {
	OUTSTR("%%+\tfont Symbol");/* need in /D and /FS for bullet symbol */
	OUTNL();
    }

    OUTSTR("%%DocumentNeededResources: (atend)");
    OUTNL();

    OUTSTR("%%DocumentSuppliedResources: (atend)");
    OUTNL();

    OUTSTR("%%Orientation: ");
    OUTSTR(landscape_mode ? "Landscape" : "Portrait");
    OUTNL();

    OUTSTR("%%PageOrder: ");
    OUTSTR(reverse ? "Descend" : "Ascend");
    OUTNL();

    OUTSTR("%%Pages: (atend)");
    OUTNL();
    OUTSTR("%%EndComments");
    OUTNL();
}

static void
#if STDC
Do_File(
FILE				*fp
)
#else /* NOT STDC */
Do_File(fp)
FILE				*fp;
#endif /* STDC */
{
    register int		c;	/* character */
    register int		last_c;	/* character */

    x = LEFT_X;
    x_right = col_width - col_margin;
    if (init_top_margin > 0)
	y = TOP_Y(init_top_margin);
    else
	y = TOP_Y(top_margin);

    Beg_Page();
    Beg_Column();

    do				/* discard all initial FF or NUL characters */
    {
	c = inchar(fp);
    }
    while ((c == (int)'\f') || (c == NUL));

    UNGETCHAR(c);		/* put back lookahead */

    for (last_c = NUL; (c = inchar(fp)) != EOF; last_c = c ? c : last_c)
    {					/* main file filter loop */
	Wrap_Check(c);

	/* We delay end-of-column action after a newline until the next */
	/* non-NUL character is seen, so as to avoid an extra empty */
	/* output page that happened when the last page was filled */
	/* to the last line.  If we are at the <FF> in a <LF><FF> */
	/* sequence, and are about to start a new column, then suppress */
	/* the action since it will happen later in Out_Formfeed(). */

	if (c && (y < BOT_Y) && !((last_c == '\n') && (c == '\f')))
	{
	    End_Column();
	    if (column > numcol)
		Beg_Page();
	    Beg_Column();
	}

	if (manpage && (Out_Manchar(fp,c) == EOF))
	    continue;	/* then character was handled in Out_Manchar() */

	switch (c)
	{
	case NUL:
	    break;	/* discard NUL's from word-padded files */

	case '\b':
	    Out_Backspace();
	    break;

	case '\f':
	    Out_Formfeed();
	    break;

	case '\v':
	    if (manpage)
		Out_Vskip(1.0);		/* reverse line feed */
	    else
		Out_Other(c);
	    break;

	case '\n':
	    Out_Newline();
	    break;

	case '\r':
	    Out_Return(fp);
	    break;

	case '\t':
	    Out_Tab();
	    break;

	case '\033':
	    Out_Escape(fp);
	    break;

	case '@':
	    Out_Atsign(fp);
	    break;

	case '_':
	    Out_Underline(fp);
	    break;

	default:
	    Out_Other(c);
	    break;
	}
    }
    column = numcol;			/* to force End_Page() action */
    End_Column();

    /* Normally, Beg_Page() sets the page_table[] entry.  Since this is
    the last output for this file, we must set it here manually.  If
    more files are printed, this entry will be overwritten (with the
    same value) at the next call to Beg_Page(). */

    page_table[page_number-init_page_number] = (long)ftell(tf);
}

static void
Do_Fonts(VOID_ARG)
{
    Read_Metrics(&normal);

    if (!normal.resident)
	normal.resident = !Download_Font(normal.name);

    if (manpage || scribe)
    {
	Read_Metrics(&bold);
	Read_Metrics(&italic);
	if (!bold.resident)
	    bold.resident = !Download_Font(bold.name);
	if (!italic.resident)
	    italic.resident = !Download_Font(italic.name);
    }

    if (scribe)
    {
	Read_Metrics(&bolditalic);
	Read_Metrics(&typewriter);
	if (!bolditalic.resident)
	    bolditalic.resident = !Download_Font(bolditalic.name);
	if (!typewriter.resident)
	    typewriter.resident = !Download_Font(typewriter.name);
    }

    if (manpage)
    {
	OUTSTR("%%IncludeResource: font Symbol");
	OUTNL();		/* need in /D and /FS for bullet symbol */
    }
}

static void
Do_Initialization(VOID_ARG)
{
#if    KCC_20
    /* Hack to make stdin and stdout avoid CRLF translation */
#if 0
    stdin->sioflgs  &= ~_SIOF_CONVERTED;
    stdin->sioflgs  |= _SIOF_UNCONVERTED;
    stdin->sioflgs  |= _SIOF_BINARY;
    stdout->sioflgs &= ~_SIOF_CONVERTED;
    stdout->sioflgs |= _SIOF_UNCONVERTED;
    stdout->sioflgs |= _SIOF_BINARY;
#endif /* 0 */

    _uioflgs[_uioufx[0]] &= ~_UIO_CONVERTED;
    _uioflgs[_uioufx[1]] &= ~_UIO_CONVERTED;
#endif /* KCC_20 */

    /* establish defaults for all parameters */
    afmpath = (const char*)getenv(AFMENVVAR); /* permit run-time override */
    if (afmpath == CNULL)
	afmpath = AFMPATH;

    baseline_skip = 0;			/* will be reset later */
    bottom_margin = DIMEN_TO_QSU(paper[PAPER_DEFAULT].bottom_margin);

    clip = FALSE;
    column = 1;
    col_margin = 0L;
    comment_line = FALSE;
    copies = 1;
    current_font = &normal;
    current_id = (int)'P';

    do_page_number = FALSE;

    init_page_number = 1;
    init_top_margin = -1;	/* negative value flags unset state */
    in_col = 0;
    in_str = 0;
    ISO8859_1_charset = TRUE;

    landscape_mode = FALSE;
    left_margin = DIMEN_TO_QSU(paper[PAPER_DEFAULT].left_margin);
    line_width = 0L;

    manpage = FALSE;
    manual_feed = 0;
    msg_prefix = "%"; /* TOPS-20 style warning message prefix for visibility */

    npages = 0;
    numcol = 1;

    out_column = 0;
    outline = 0L;
    output_this_page = TRUE;

    page_count = 0;
    page_height = DIMEN_TO_QSU(paper[PAPER_DEFAULT].height);
    page_number = init_page_number;
    page_width = DIMEN_TO_QSU(paper[PAPER_DEFAULT].width);
    pages = (PAGE_SELECTION*)NULL;
    paper_name = paper[PAPER_DEFAULT].name;
    point_size = DIMEN_TO_QSU("10bp");

    quiet = FALSE;

    reverse = FALSE;	/* used to be TRUE, but changed at 3.0.0 to FALSE */
    right_margin = DIMEN_TO_QSU(paper[PAPER_DEFAULT].right_margin);

    scribe = FALSE;

    top_margin = DIMEN_TO_QSU(paper[PAPER_DEFAULT].top_margin);

    wrap = FALSE;

    normal.name = ps_normal[PS_DEFAULT];
    bold.name = ps_bold[PS_DEFAULT];
    italic.name = ps_italic[PS_DEFAULT];
    typewriter.name = ps_typewriter[PS_DEFAULT];
    bolditalic.name = ps_both[PS_DEFAULT];

    normal.resident = TRUE;		/* default fonts are printer-resident */
    bold.resident = TRUE;
    italic.resident = TRUE;
    typewriter.resident = TRUE;
    bolditalic.resident = TRUE;

    Do_Startup_File();
    errno = 0;				/* clear global error flag */
}

static void
Do_ISO8859_1_encoding(VOID_ARG)
{
    static const char* encoding[] =
    {
	"/ISOLatin1Encoding where {pop save true}{false} ifelse",
	"/ISOLatin1Encoding [",
	"    /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef",
	"    /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef",
	"    /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef",
	"    /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef",
	"    /.notdef /.notdef /.notdef /.notdef /.notdef /.notdef",
	"    /.notdef /.notdef /space /exclam /quotedbl /numbersign",
	"    /dollar /percent /ampersand /quoteright /parenleft",
	"    /parenright /asterisk /plus /comma /minus /period",
	"    /slash /zero /one /two /three /four /five /six /seven",
	"    /eight /nine /colon /semicolon /less /equal /greater",
	"    /question /at /A /B /C /D /E /F /G /H /I /J /K /L /M",
	"    /N /O /P /Q /R /S /T /U /V /W /X /Y /Z /bracketleft",
	"    /backslash /bracketright /asciicircum /underscore",
	"    /quoteleft /a /b /c /d /e /f /g /h /i /j /k /l /m",
	"    /n /o /p /q /r /s /t /u /v /w /x /y /z /braceleft",
	"    /bar /braceright /asciitilde /guilsinglright /fraction",
	"    /florin /quotesingle /quotedblleft /guilsinglleft /fi",
	"    /fl /endash /dagger /daggerdbl /bullet /quotesinglbase",
	"    /quotedblbase /quotedblright /ellipsis /trademark",
	"    /perthousand /grave /scaron /circumflex /Scaron /tilde",
	"    /breve /zcaron /dotaccent /dotlessi /Zcaron /ring",
	"    /hungarumlaut /ogonek /caron /emdash /space /exclamdown",
	"    /cent /sterling /currency /yen /brokenbar /section",
	"    /dieresis /copyright /ordfeminine /guillemotleft",
	"    /logicalnot /hyphen /registered /macron /degree",
	"    /plusminus /twosuperior /threesuperior /acute /mu",
	"    /paragraph /periodcentered /cedilla /onesuperior",
	"    /ordmasculine /guillemotright /onequarter /onehalf",
	"    /threequarters /questiondown /Agrave /Aacute",
	"    /Acircumflex /Atilde /Adieresis /Aring /AE /Ccedilla",
	"    /Egrave /Eacute /Ecircumflex /Edieresis /Igrave /Iacute",
	"    /Icircumflex /Idieresis /Eth /Ntilde /Ograve /Oacute",
	"    /Ocircumflex /Otilde /Odieresis /multiply /Oslash",
	"    /Ugrave /Uacute /Ucircumflex /Udieresis /Yacute /Thorn",
	"    /germandbls /agrave /aacute /acircumflex /atilde",
	"    /adieresis /aring /ae /ccedilla /egrave /eacute",
	"    /ecircumflex /edieresis /igrave /iacute /icircumflex",
	"    /idieresis /eth /ntilde /ograve /oacute /ocircumflex",
	"    /otilde /odieresis /divide /oslash /ugrave /uacute",
	"    /ucircumflex /udieresis /yacute /thorn /ydieresis",
	"] def {restore} if",
	(const char *)NULL
    };

    Out_Lines(tf,encoding);
    Reencode();
}

static void
#if STDC
Do_Options(
int				argc,
char				*argv[]
)
#else /* NOT STDC */
Do_Options(argc,argv)
int				argc;
char				*argv[];
#endif /* STDC */
{
    register int		k;	/* loop index */

    bad_option = FALSE;

    for (k = 1; k < argc; ++k)
    {
	if (strchr(OPTIONCHARS,argv[k][0]) != (char *)NULL)
	    Option((const char*)argv[k]);
    }

    if (bad_option)
	Usage(EXIT_FAILURE);		/* never returns */
}

static void
Do_PageSetup(VOID_ARG)
{
    static const char *page_setup[] =
    {
	"%%BeginPageSetup",
	"lptopsdict begin",
	"PB",
	"%%EndPageSetup",
	CNULL
    };

    Out_Lines(tf,page_setup);
}

static void
Do_PageTable(VOID_ARG)
{
    register int		k;	/* loop index */
    register int		n;	/* pages output */

    tf = stdout;			/* so OUTSTR() macro works */

    PRINTF("%%%%Pages: %d",page_count);
    NEWLINE(stdout);

    OUTSTR("%%PageTable: ");
    for (k = 0, n = 0; k < MIN(MAXPAGE,page_number-init_page_number+1); ++k)
    {
	Check_Page_Selection(k + init_page_number);
	if ((n > 0) && ((n % 4) == 0))
	{
	    OUTNL();
	    OUTSTR("%%+");
	}
	if (output_this_page)
	    n++;
	SPRINTF(tempstr,"\t%d %d %ld",init_page_number+k,init_page_number+k,
	    page_table[k]);
	OUTSTR(tempstr);
    }
    output_this_page = TRUE;
    OUTNL();
}

static void
Do_Prolog(VOID_ARG)
{
    static const char *prolog_1[] =
    {
	"%%BeginProlog",
	"/lptopsdict 50 dict def",
	"lptopsdict begin",
	"/CB {gsave mark} bind def % Column Begin",
	"/CE {cleartomark grestore} bind def % Column End",
	"/COLUMNPATH { % x y COLUMNPATH",
	CNULL
    };

    static const char *prolog_2[] =
    {
	"/E {grestore 0 BASELINESKIP neg rmoveto} bind def % End line",
	CNULL
    };

    static const char *prolog_3[] =
    {
	"/O {gsave show grestore} bind def % Overprint",
	"/P {  % Print with newline",
	"\tgsave show grestore 0 BASELINESKIP neg rmoveto",
	"} bind def",
	"/R {grestore} bind def",
	"/S {gsave} bind def",
	"/W {show} bind def % Show text",
	"/PB { % Page Begin",
	"\t% Define image orientation: one of PORTRAIT, LANDSCAPE,",
	"\t% ROTATE_0, ROTATE_90, ROTATE_180, or ROTATE_270",
	CNULL
    };

    static const char *prolog_4[] =
    {
	"\t/SavePage save def",
	"} bind def",
	"/PE { % Page End",
	"\tSavePage restore",
	"} bind def",
	"/LANDSCAPE {ROTATE_90} bind def",
	"/PORTRAIT {} def",
	"/ROTATE_0 {} def",
	"/ROTATE_90 {HEIGHT 0 translate 90 rotate} bind def",
	"/ROTATE_180 {WIDTH HEIGHT translate 180 rotate} bind def",
	"/ROTATE_270 {0 WIDTH translate 270 rotate} bind def",
	"/U { % (char) U -- show underlined character",
	"\tcurrentpoint /Y exch def /X exch def",
	"\tdup FN stringwidth pop",
	"\texch show",
	"\tgsave",
	/* Examination of the 61 *.afm files in /usr/openwin/lib/fonts/afm
	   on Sun Solaris 2.1 found underscore thicknesses ranging from
	   41 to 111, with an average of 57, and most at 50.  The
	   lower-left corner position varied from -252 to -60, with an
	   average of -129, and most at -125.  These are on a pointsize
	   scale of 1000, so we use 0.050 * POINTSIZE for the underscore
	   line width, and set its lower-left corner -0.125 * POINTSIZE
	   below the baseline, or equivalently, the midpoint of its left
	   end at -0.100 below the baseline, with line butt caps flush
	   with the start and end points.*/
	"\t\tPOINTSIZE 0.050 mul setlinewidth",
	"\t\t0 setlinecap",
	"\t\tnewpath",
	"\t\t\tX Y moveto",
	"\t\t\t0 POINTSIZE -0.100 mul rmoveto",
	"\t\t\t0 rlineto",
	"\t\tstroke",
	"\tgrestore",
	"} bind def",
	CNULL
    };

    static const char *prolog_5[] =
    {
	"/D {FS (\\267) show} bind def % bullet",
	"/FS {/Symbol findfont POINTSIZE scalefont setfont} bind def",
	"/V {BASELINESKIP mul 0 exch rmoveto} def % n V -- vertical skip",
	"/u {0.5 V} def",
	"/d {-0.5 V} def",
	CNULL
    };

    static const char *prolog_6[] =
    {
	"/MF {",
	"\tstatusdict begin",
	"\t\tstatusdict /manualfeed true put",
	"\tend",
	"} def",
	CNULL
    };

    static const char *prolog_7[] =
    {
	"end % lptopsdict",
	"%%EndProlog",
	CNULL
    };

    Do_Separator();

    Out_Lines(tf,prolog_1);

    if (outline > 0)
    {
	FPRINTF(tf, "\t%ld setlinewidth", outline);
	OUTNL();
    }
    FPRINTF(tf,"\tnewpath moveto");
    OUTNL();
    FPRINTF(tf, "\t%ld %ld rlineto", (long)col_width, 0L);
    OUTNL();
    FPRINTF(tf, "\t%ld %ld rlineto",
	0L, (long)(page_height-bottom_margin-top_margin));
    OUTNL();
    FPRINTF(tf, "\t%ld %ld rlineto", -(long)col_width, 0L);
    OUTNL();
    FPRINTF(tf, "\tclosepath");
    OUTNL();
    FPRINTF(tf, "} bind def");
    OUTNL();

    FPRINTF(tf,"/NEWCOLUMN { COLUMNPATH%s%s } bind def",
	(outline > 0) ? " gsave stroke grestore" : "",
	clip ? " clip" : "");
    OUTNL();

    FPRINTF(tf,"/POINTSIZE %ld def %% %gbp == %gpt",
	point_size,QSU_TO_BP(point_size),QSU_TO_PT(point_size));
    OUTNL();

    if (manpage)
    {
	/* bullet width is constant with this entry in Symbol.afm: */
	/* Comment Creation Date: Thu Sep 28 03:29:45 1989 */
	/* Version 001.005 */
	/* C 183 ; WX 460 ; N bullet ; B 50 113 410 473 ; */
#define BULLETWIDTH	((point_size * 460) / 1000L)
	FPRINTF(tf,"/BULLETWIDTH {POINTSIZE 460 mul 1000 div} bind def");
	OUTNL();
    }

    FPRINTF(tf,"/BASELINESKIP %ld def %% %gbp == %gpt",
	baseline_skip,QSU_TO_BP(baseline_skip),QSU_TO_PT(baseline_skip));
    OUTNL();

    OUTSTR("/HEIGHT ");
    FPRINTF(tf, "%g def %% %gin == %gcm", QSU_TO_BP(page_height),
	QSU_TO_IN(page_height),QSU_TO_CM(page_height));
    OUTNL();

    OUTSTR("/WIDTH ");
    FPRINTF(tf, "%g def %% %gin == %gcm", QSU_TO_BP(page_width),
	QSU_TO_IN(page_width),QSU_TO_CM(page_width));
    OUTNL();

    if (ISO8859_1_charset)
	Do_ISO8859_1_encoding();

    if (manpage || scribe)
    {
	Out_Fontdef("B", bold.name);
	Out_Fontdef("I", italic.name);
	Out_Fontdef("N", normal.name);
	if (scribe)
	{
	    Out_Fontdef("H", bolditalic.name);
	    Out_Fontdef("T", typewriter.name);
	}
	Out_Lines(tf,prolog_2);
    }
    else
	Out_Fontdef("N", normal.name);
    Out_Lines(tf,prolog_3);

    OUTSTR(landscape_mode ? "\tLANDSCAPE" : "\tPORTRAIT");
    OUTNL();

    FPRINTF(tf,"\t72 %ld div dup scale",QSU);
    OUTNL();

    Out_Lines(tf,prolog_4);

    if (manpage)
	Out_Lines(tf,prolog_5);

    if (do_page_number)
    {
	OUTSTR("/PN { % (string) PN -- display pagenumber");
	OUTNL();

	OUTSTR("\tFN ");

	switch (page_flag)
	{
	case (PAGE_T | PAGE_L):
	    FPRINTF(tf,"%ld %ld moveto show",left_margin,
		page_height-top_margin/2-baseline_skip/2);
	    break;

	case (PAGE_T | PAGE_C):
	    FPRINTF(tf,
		"%ld %ld moveto dup stringwidth pop 2 div neg 0 rmoveto show",
		page_width/2, page_height-top_margin/2-baseline_skip/2);
	    break;

	case (PAGE_T | PAGE_R):
	    FPRINTF(tf,
		"%ld %ld moveto dup stringwidth pop neg 0 rmoveto show",
		page_width-left_margin,
		page_height-top_margin/2-baseline_skip/2);
	    break;

	case (PAGE_B | PAGE_L):
	    FPRINTF(tf,"%ld %ld moveto show",
		left_margin, bottom_margin/2-baseline_skip/2);
	    break;

	case (PAGE_B | PAGE_C):
	    FPRINTF(tf,
		"%ld %ld moveto dup stringwidth pop 2 div neg 0 rmoveto show",
		page_width/2, bottom_margin/2-baseline_skip/2);
	    break;

	case (PAGE_B | PAGE_R):
	    FPRINTF(tf,
		"%ld %ld moveto dup stringwidth pop neg 0 rmoveto show",
		page_width-left_margin, bottom_margin/2-baseline_skip/2);
	    break;

	default:			/* else no numbering */
	    OUTSTR("pop");
	    break;
	}
	OUTNL();

	OUTSTR("} bind def");
	OUTNL();
    }

    if (manual_feed > 0)
	Out_Lines(tf,prolog_6);

    Out_Lines(tf,prolog_7);

    Do_Separator();
}

static void
Do_Startup_File(VOID_ARG)
{
    const char			*initpath;	/* initfile search path */
    FILE			*initfile;
    char			*initfilename;
    char			line[4*BUFSIZ];	/* allow long input lines */
    char			*p;	/* pointer into line[] */

    initpath = (const char*)getenv(INITENVVAR);
    initfilename = findfile((initpath == (const char *)NULL) ? INITPATH :
	initpath, INITFILE);

    if ((initfilename == (char*)NULL) ||
	(initfile = fopen(initfilename, FOPEN_R)) == (FILE*)NULL)
	return;	/* no initialization file found, so quietly ignore failure */

    while (Get_Wrapped_Line(line,sizeof(line),initfile) != (char*)NULL)
    {
	Strip_Comments(line,INITFILECOMMENTCHAR);
	for (p = &line[0]; *p && isspace(*p); ++p)
	    /* NO-OP */;		/* skip leading whitespace */
	if (*p == NUL)
	    /* NO-OP */;		/* ignore empty lines */
	else if (strchr(OPTIONCHARS,*p) != (char*)NULL)
	    Option((const char*)p);
	else
	    Opt_Error(CNULL,(const char*)p);
    }
    (void)fclose(initfile);
}

static void
Do_Resources(VOID_ARG)
{
#define MAXFAMILY	6		/* normal, bold, italic, bolditalic,
					typewriter, and symbol */
    int				k;	/* loop index */
    const char			*font_downloaded[MAXFAMILY];
    const char			*font_resident[MAXFAMILY];
    int				n_resident;
    int				n_downloaded;

#define RECORD_FONT_USE(f)	if (f.resident) \
	font_resident[n_resident++] = f.name; \
    else \
	font_downloaded[n_downloaded++] = f.name

    n_resident = 0;
    n_downloaded = 0;
    RECORD_FONT_USE(normal);
    if (manpage || scribe)
    {
	RECORD_FONT_USE(bold);
	RECORD_FONT_USE(italic);
    }
    if (scribe)
    {
	RECORD_FONT_USE(bolditalic);
	RECORD_FONT_USE(typewriter);
    }
    if (manpage)		/* need in /D and /FS for bullet symbol */
	font_resident[n_resident++] = "Symbol";

    /* Both of these are referred to by (atend) declarations in the preamble, */
    /* so both must be present, even if their font lists are empty. */

    OUTSTR("%%DocumentNeededResources: ");
    for (k = 0; k < n_resident; ++k)
	Out_Font_Resource(font_resident[k],k);
    OUTNL();

    OUTSTR("%%DocumentSuppliedResources: ");
    for (k = 0; k < n_downloaded; ++k)
	Out_Font_Resource(font_downloaded[k],k);
    OUTNL();
}

static void
#if STDC
Do_Scribe_string(
FILE *fp,
int c
)
#else
Do_Scribe_string(fp,c)
FILE *fp;
int c;
#endif
{
    int				save_id;
    Font			*save_font;

    End_Text();		/* complete any open string before font change */
    save_id = current_id;
    save_font = current_font;
    current_id = c;
    switch (c)
    {
    case 'B':	current_font = &bold; break;
    case 'H':	current_font = &bolditalic; break;
    case 'I':	current_font = &italic; break;
    case 'T':	current_font = &typewriter; break;
    default:	current_font = &normal; break;
    }
    Copy_Plist(fp);
    current_id = save_id;
    current_font = save_font;
}

static void
Do_Separator(VOID_ARG)
{
    OUTSTR("%===================================");
    OUTSTR("====================================");
    OUTNL();
}

static void
Do_Setup(VOID_ARG)
{
    OUTSTR("%%BeginSetup");
    OUTNL();

    FPRINTF(tf,"/#copies %d def",copies);
    OUTNL();

    Do_Fonts();

    OUTSTR("%%EndSetup");
    OUTNL();

    Do_Separator();
}

static void
Do_Tempfile(VOID_ARG)
{
    if (reverse)	/* then create initial output in temporary file */
    {
#if OS_UNIX
	if ((tf = Fopen_Tempfile(TEMPFILE_1,FOPEN_RW)) == (FILE *)NULL)
	    Fatal("Cannot open temporary file [%s]",TEMPFILE_1);
	unlink(TEMPFILE_1); /* so temp file disappears even if we abort */
#else
	if ((tf = Fopen_Tempfile(TEMPFILE_1,FOPEN_W)) == (FILE *)NULL)
	    Fatal("Cannot open temporary file [%s]",TEMPFILE_1);
#endif
    }
    else
	tf = stdout;
    errno = 0;				/* clear global error flag */
}

static void
Do_Trailer(VOID_ARG)
{
    tf = stdout;			/* so OUTSTR() macro works */
    output_this_page = TRUE;

    OUTSTR("%%Trailer");
    NEWLINE(stdout);

    Do_Resources();
    Do_PageTable();

#if __hppa
    /* HP-UX 9.0 has a bug that results in bad return values
       from [a-z]*printf() if ftell() has ever been called.  We
       therefore ignore the return code, and only check errno here. */
    FPRINTF(tf,"%%%%EOF");
#else
    if (fprintf(tf,"%%%%EOF") == EOF)
	Fatal("Output error---disk storage probably full",CNULL);
#endif

    OUTNL();

    if (!quiet)
    {
	FPRINTF(stderr," [ok]");
	NEWLINE(stderr);
    }
}

static void
Do_Validation(VOID_ARG)
{					/* validate input settings */
    copies = MAX(1,copies);
    numcol = MAX(1,numcol);
    point_size = MAX(DIMEN_TO_QSU("3bp"),point_size);
    if ((QSU_TO_BP(point_size) > 18.0) && !quiet)
	Warning("Font point size > 18bp is unusually large!", CNULL);
    outline = MAX(0L,outline);

    /* TeX's baselineskip is 12pt for 10pt type; we preserve the same ratio */
    if (baseline_skip < point_size)
	baseline_skip = (point_size * 12L) / 10L;

    if (landscape_mode && (page_width < page_height))
    {	/* swap sizes for landscape mode, unless paper setting already did so */
	COORDINATE swap = page_height;

	page_height = page_width;
	page_width = swap;
    }

    if (((left_margin + right_margin) >= page_width) ||
	((top_margin + bottom_margin) >= page_height) ||
	((init_top_margin + bottom_margin) >= page_height))
	Fatal("Margin values too large for page",CNULL);

    col_width = (page_width - left_margin - right_margin)/numcol;
    col_margin = ((numcol == 1) && (outline == 0)) ?
	0L : (col_width/COL_TWIDDLE);
}

static BOOLEAN
#if STDC
Download_Font(
const char			*font_name
)
#else /* NOT STDC */
Download_Font(font_name)
const char			*font_name;
#endif /* STDC */
{					/* return TRUE if font was downloaded */
    char			the_basename[MAXPATHLEN+1]; /* font filename */
    int				c;
    char			*filename;
    FILE			*fpin;	/* font file handle */

    SPRINTF(the_basename,"%s%s",font_name,PFAEXT);
    filename = findfile(afmpath, the_basename); /* first try ASCII font file */

    if (filename == (char*)NULL)	/* then try map file */
	filename = Map_Font(font_name,PFAEXT);

    if (filename == (char*)NULL)	/* then try binary font file */
    {
	SPRINTF(the_basename,"%s%s",font_name,PFBEXT);
	filename = findfile(afmpath, the_basename);
    }

    if (filename == (char*)NULL)	/* then try map file */
	filename = Map_Font(font_name,PFBEXT);

    fpin = (filename == (char*)NULL) ? (FILE*)NULL : fopen(filename,FOPEN_R);

    if (fpin == (FILE*)NULL)		/* no downloadable font available */
    {
	if (filename != (char*)NULL)
	    Warning("Cannot open font file [%s]",filename);
	OUTSTR("%%IncludeResource: font ");
	OUTSTR(font_name);
	OUTNL();
	return (FALSE);
    }
    else				/* download the font */
    {
	OUTSTR("%%BeginResource: font ");
	OUTSTR(font_name);
	OUTNL();

	OUTSTR("% file: ");
	OUTSTR(filename);
	OUTNL();

	c = ungetc(fgetc(fpin),fpin);	/* peek at first character */
	if (c == PFB_MAGIC)		/* magic signature of .pfb file */
	    Copy_PFB(tf, fpin);
	else				/* assume normal .pfa (text) file */
	    Copy_PFA(tf, fpin);
	OUTNL();

	OUTSTR("% font: ");
	OUTSTR(font_name);
	OUTNL();

	OUTSTR("% file: ");
	OUTSTR(filename);
	OUTNL();

	OUTSTR("%%EndResource");
	OUTNL();

	(void)fclose(fpin);
	return (TRUE);
    }
}

static void
#if STDC
Dump_Font(
const char* ps_name,
const char* ps_fonts[]
)
#else /* NOT STDC */
Dump_Font(ps_name,ps_fonts)
const char* ps_name;
const char* ps_fonts[];
#endif /* STDC */
{
    register int		m;	/* loop index */

    NEWLINE(stderr);
    FPRINTF(stderr,ps_name);
    NEWLINE(stderr);

    for (m = 0; ps_fonts[m] != CNULL; m += 4)
    {
	if ((m > 3) &&
	    (strcmp(ps_fonts[m],ps_fonts[m-4]) == 0) &&
	    (strcmp(ps_fonts[m+1],ps_fonts[m+1-4]) == 0) &&
	    (strcmp(ps_fonts[m+2],ps_fonts[m+2-4]) == 0) &&
	    (strcmp(ps_fonts[m+3],ps_fonts[m+3-4]) == 0))
	    continue;			/* don't show duplicate lines */
	FPRINTF(stderr,"\t%-30s %-10s %-10s %-10s",
	    ps_fonts[m], ps_fonts[m+1], ps_fonts[m+2], ps_fonts[m+3]);
	NEWLINE(stderr);
    }
}

static void
Dump_Paper(VOID_ARG)
{
    int				k;	/* loop index */

#define PAPERFMT	"%-22s %8s %8s %8s %8s %8s %8s"

    NEWLINE(stderr);
    FPRINTF(stderr,"Known paper sizes:");
    NEWLINE(stderr);
    NEWLINE(stderr);
    FPRINTF(stderr,PAPERFMT,
	"Name",
	"width",
	"height",
	"left",
	"right",
	"bottom",
	"top");
    NEWLINE(stderr);

    for (k = 0; paper[k].name != CNULL; ++k)
    {
	FPRINTF(stderr,PAPERFMT,
	    paper[k].name,
	    paper[k].width,
	    paper[k].height,
	    paper[k].left_margin,
	    paper[k].right_margin,
	    paper[k].bottom_margin,
	    paper[k].top_margin);
	NEWLINE(stderr);
    }
}

static void
End_Column(VOID_ARG)
{
    End_Line();
    OUTSTR("CE");
    OUTNL();
    if (!quiet && output_this_page)
	FPRINTF(stderr,
	    ".%d",line_number);	/* continue [p.l.l...] status report */
    if (++column > numcol)	/* columns filled, so start new page */
    {
	if (!quiet && output_this_page)
	    (void)putc(']',stderr);/* close [p.l.l...] report */
	End_Page();
    }
    else			/* just start new column */
    {
	x += col_width;
	y = TOP_Y(top_margin);
    }
}

static void
End_Page(VOID_ARG)
{

#ifdef FORMFEED
    register int		k;
#endif /* FORMFEED */

    if (do_page_number)
    {
	if ((init_top_margin > 0) && (page_number == init_page_number)
	    && (page_flag | PAGE_T))
	    ;			/* omit leading top pagenumber */
	else if (page_flag & PAGE_C)
	{
	    SPRINTF(tempstr,"(- %d -) PN",page_number);
	    OUTSTR(tempstr);
	    OUTNL();
	}
	else
	{
	    SPRINTF(tempstr,"(%d) PN",page_number);
	    OUTSTR(tempstr);
	    OUTNL();
	}
    }
    page_number++;

    if (manual_feed > 0)
    {
	OUTSTR("MF");
	OUTNL();
	manual_feed--;
    }

    OUTSTR("PE");
    OUTNL();

    OUTSTR("end % lptopsdict");
    OUTNL();

    OUTSTR("showpage");
    OUTNL();

    OUTSTR("%%PageTrailer");
    OUTNL();
    if (DISKFULL(tf))
	Fatal("Output error---disk storage probably full",CNULL);

    Do_Separator();

#ifdef FORMFEED
    for (k = 1; k <= copies; ++k)
	OUTCHAR('\f'); /* output FF's for print spooler page accounting */
#endif /* FORMFEED */

    x = LEFT_X;
    y = TOP_Y(top_margin);
    if (output_this_page)
	page_count++;
}

static void
End_Line(VOID_ARG)
{
    if (BOL())
	return;			/* no line to end */
    End_Text();
    OUTCHAR(' ');
    if (manpage || scribe)
	OUTCHAR('E');	/* grestore in E matches gsave from Beg_Line() */
    if (out_column > 0)		/* start new line, unless we just did */
	OUTNL();
}

static void
End_String(VOID_ARG)
{
    while (level > 0)
    {
	OUTCHAR(')');
	level--;
    }
    in_str = 0;
}

static void
End_Text(VOID_ARG)
{
    int				old_level = level;

    End_String();
    if (old_level > 0)
    {
	if (out_column >= (MAXLINEWIDTH-1)) /* can wrap after string */
	    OUTNL();
	OUTCHAR(current_id);
    }
}

static void
#if STDC
Fatal(
const char			*fmt,
const char			*arg
)
#else /* NOT STDC */
Fatal(fmt,arg)
const char			*fmt;
const char			*arg;
#endif
{
    msg_prefix = "?";		/* fatal error prefix */
    Warning(fmt,arg);
    exit(EXIT_FAILURE);
}

static FILE *
#if STDC		
Fopen_Tempfile(
char				*filename,
const char			*mode
)
#else
Fopen_Tempfile(filename, mode)
char				*filename;
const char			*mode;
#endif
{
    /* Supply a temporary filename if the filename is empty, and then
       return an open file pointer to it. */

#if defined(HAVE_MKSTEMP) && defined(HAVE_MKSTEMP_DECL) && defined(OS_UNIX)
    int fd;
    if (filename[0] == '\0')
	(void)strcpy(filename, DEFAULT_TMPFILE);
    fd = mkstemp(filename);
    return ((fd == -1) ? (FILE*)NULL : fdopen(fd, mode));
#else
    return (fopen(filename, mode));
#endif        
}

static int
#if STDC
Get_Bold_Char(
FILE *fp
)
#else /* NOT STDC */
Get_Bold_Char(fp)
FILE *fp;
#endif /* STDC */
{				/* return c from c(<BS>c)+, or EOF if none */
    int				c;
    int				overstrikes;
    int				next_c;
    int				next_next_c;

    c = inchar(fp);		/* peek at 3 successive characters */
    next_c = inchar(fp);
    next_next_c = inchar(fp);

    for (overstrikes = 0;
	(next_c == (int)'\b') && (next_next_c == c); ++overstrikes)
    {				/* gobble overstrikes */
	next_c = inchar(fp);
	next_next_c = inchar(fp);
    }
    UNGETCHAR(next_next_c);	/* put back two-character lookahead */
    UNGETCHAR(next_c);

    if (overstrikes == 0)	/* failure: no overstrikes found */
    {
	UNGETCHAR(c);	/* put back first lookahead character */
	return (EOF);
    }
    else			/* success */
	return (c);
}

static int
#if STDC
Get_Underlined_Char(
FILE *fp
)
#else /* NOT STDC */
Get_Underlined_Char(fp)
FILE *fp;
#endif /* STDC */
{				/* return c from _<BS>c, or EOF if none */
    int				c;
    int				next_c;

    c = inchar(fp);		/* peek at next character */
    if (c != (int)'_')		/* then no underlined character */
    {
	UNGETCHAR(c);
	return (EOF);
    }

    next_c = inchar(fp);	/* peek at second following character */
    if (next_c != (int)'\b')	/* then no underlined character */
    {
	UNGETCHAR(next_c);
	UNGETCHAR(c);
	return (EOF);
    }

    return (inchar(fp));	/* found underlined character */
}

static char*
#if STDC
Get_Wrapped_Line(
char				*line,
size_t				maxline,
FILE				*fpin
)
#else /* NOT STDC */
Get_Wrapped_Line(line,maxline,fpin)
char				*line;
size_t				maxline;
FILE				*fpin;
#endif /* STDC */
{
    size_t			n;	/* how many bytes already input */

    for (n = 0; n < maxline; )
    {
	if (fgets(line + n,(int)(maxline - n),fpin) == (char*)NULL)
	    break;			/* reached end of file */
	n = strlen(line);
	if ((n >= 2) && (line[n-2] == '\\') && (line[n-1] == '\n'))
	    n -= 2;			/* have partial wrapped line */
	else
	    break;			/* have complete line */
    }
    return ((n > 0) ? line : (char*)NULL);
}

static int
#if STDC
inchar(
FILE *fp
)
#else /* NOT STDC */
inchar(fp)
FILE *fp;
#endif /* STDC */
{
    register int		c;
    static int			lastc = NUL;
				/* memory of previous input character */

    /* Output is incorrect if the file does not end with a newline;
    this routine supplies one at EOF if it is lacking.

    On UNIX, lptops is often used with "pr -f", which ends its output
    with <newline><formfeed> (Berkeley) or just <formfeed> (AT&T).  We
    handle this case specially by discarding the formfeed; otherwise,
    the default action of supplying a linefeed before EOF will produce
    an empty page in the output of lptops. */

    c = GETC(fp);
    if ((lastc == (int)'\n') &&
	(c == (int)'\f') &&
	(ungetchar(GETC(fp)) == EOF))
	return (EOF);		/* Berkeley-style <newline><formfeed><EOF> */
    else if ((c == (int)'\f') && (ungetchar(GETC(fp)) == EOF))
	c = (int)'\n';		/* AT&T style <formfeed><EOF> */
    if ((c == EOF) && (lastc != (int)'\n'))
	c = (int)'\n';
    lastc = c;
    return (c);
}

static char *
#if STDC
Map_Font(
const char 			*font_name,
const char 			*ext
)
#else /* NOT STDC */
Map_Font(font_name,ext)
const char 			*font_name;
const char 			*ext;
#endif /* STDC */
{
    char			line[2*MAXPATHLEN+1]; /* mapfile line buffer */
    FILE			*mapfile; /* AFM mapfile handle */
    char			*mapfilename; /* filename in mapfile */
    char			*mapfont; /* fontname in mapfile */
    char			*p;	/* pointer into mapfilename */

    mapfilename = findfile(afmpath, AFMMAPFILE);

    if ((mapfile = fopen(mapfilename, FOPEN_R)) == (FILE*)NULL)
	return ((char*)NULL);		/* no mapfile found */

    while (Get_Wrapped_Line(line,sizeof(line),mapfile) != (char*)NULL)
    {
	Strip_Comments(line,AFMMAPCOMMENTCHAR);
	if ((mapfont = strtok(line,AFMMAPSEPCHARS)) == (char*)NULL)
	    continue;			/* no font on this line */

	if (strcmp(mapfont,font_name) == 0) /* then found the requested font */
	{
	    if ((mapfilename = strtok((char *)NULL,AFMMAPSEPCHARS)) ==
		(char*)NULL)
		continue;		/* no file for this font */
	    (void)fclose(mapfile);
	    p = strrchr(mapfilename,'.'); /* do we have a file extension? */
	    (void)strcpy((p == (char*)NULL) ?
		mapfilename + strlen(mapfilename) : p,ext);
	    return (findfile(afmpath, mapfilename));
	}
    }
    (void)fclose(mapfile);
    return ((char*)NULL);
}

static void
#if STDC
Option(
const char			*opt
)
#else /* NOT STDC */
Option(opt)
const char			*opt;
#endif /* STDC */
{
    int				k;	/* index into Option_Table[] */
    size_t			optlen;
    const char			*value;

#define ISNUMBER(c)	(isdigit(c) || ((c) == '-') || ((c) == '+'))
#define SETBOOLEAN(v)	(((int)strtol(v,(char**)NULL,0) == 0) ? FALSE : TRUE)
#define SETINT(v)	((int)strtol(v,(char**)NULL,0))

    if (strchr(OPTIONCHARS,opt[1]) != (char*)NULL)
	opt++;		/* allow GNU- and POSIX-style --longname options */

    for (value = &opt[1]; *value && isalpha(*value); ++value)
	/* NO-OP */;			/* skip over option name */

    optlen = (size_t)(value - &opt[1]);	/* remember option length */

    for ( ; *value && isspace(*value); ++value)
	/* NO-OP */;			/* skip leading space in option value */

    if ((*value == '=') || (*value == ':'))
    {					/* skip optional assignment separator */
	for (value++; *value && isspace(*value); ++value)
	    /* NO-OP */;		/* skip leading space in option value */
    }

    for (k = 0; Option_Table[k].name != CNULL; ++k)
    {
	if (strnicmp(Option_Table[k].name,&opt[1],
	    MAX(optlen,Option_Table[k].min_match)) == 0)
	{
	    switch (Option_Table[k].type)
	    {
	    case T_INT:
		if (ISNUMBER(*value))
		    *((int*)(Option_Table[k].variable)) = SETINT(value);
		else if ((*value == NUL) &&
		    (Option_Table[k].default_value != CNULL))
		    *((int*)(Option_Table[k].variable)) =
			SETINT(Option_Table[k].default_value);
		else
		    Opt_Error(CNULL,opt);
		break;

	    case T_BOOLEAN:
		if (ISNUMBER(*value))
		    *((BOOLEAN*)(Option_Table[k].variable)) = SETBOOLEAN(value);
		else if (*value == NUL)
		    *((BOOLEAN*)(Option_Table[k].variable)) =
			!*((BOOLEAN*)(Option_Table[k].variable));
		else if (strnicmp(value,"yes",strlen(value)) == 0)
		    *((BOOLEAN*)(Option_Table[k].variable)) = TRUE;
		else if (strnicmp(value,"no",strlen(value)) == 0)
		    *((BOOLEAN*)(Option_Table[k].variable)) = FALSE;
		else if (strnicmp(value,"true",strlen(value)) == 0)
		    *((BOOLEAN*)(Option_Table[k].variable)) = TRUE;
		else if (strnicmp(value,"false",strlen(value)) == 0)
		    *((BOOLEAN*)(Option_Table[k].variable)) = FALSE;
		else
		    Opt_Error(CNULL,opt);
		break;

	    case T_DIMEN:
		if (ISNUMBER(*value))
		    *((COORDINATE*)(Option_Table[k].variable)) =
			DIMEN_TO_QSU(value);
		else if ((*value == NUL) &&
		    (Option_Table[k].default_value != CNULL))
		    *((COORDINATE*)(Option_Table[k].variable)) =
			DIMEN_TO_QSU(Option_Table[k].default_value);
		else
		    Opt_Error(CNULL,opt);
		break;

	    case T_SPECIAL:
		(*(Option_Table[k].special))(opt,value);
		break;

	    default:
		Fatal("Option_Table[] corrupted",CNULL);
	    }
	    return;
	}
    }
    Opt_Error(CNULL,opt);
}

static void
#if STDC
Opt_Author(
const char			*option,
const char			*value
)
#else /* NOT STDC */
Opt_Author(option,value)
const char			*option;
const char			*value;
#endif /* STDC */
{
    static const char *author[] =
    {
	"Author:",
	"\tNelson H. F. Beebe",
	"\tCenter for Scientific Computing",
	"\tDepartment of Mathematics, LCB 110",
	"\t155 S 1400 E RM 233",
	"\tUniversity of Utah",
	"\tSalt Lake City, UT 84112-0090",
	"\tUSA",
	"\tTel: +1 801 581 5254",
	"\tFAX: +1 801 581 4148",
	"\tEmail: beebe@math.utah.edu, beebe@acm.org",
	"\t       beebe@computer.org (Internet)",
	"\tWWW URL: http://www.math.utah.edu/~beebe",			
	CNULL
    };

    Out_Lines(stderr,author);
    exit(EXIT_SUCCESS);
}

static void
#if STDC
Opt_Copyright(
const char			*option,
const char			*value
)
#else /* NOT STDC */
Opt_Copyright(option,value)
const char			*option;
const char			*value;
#endif /* STDC */
{
    static const char *copyright[] =
    {
	"\tlptops is licensed under the GNU General Public License as published by",
	"\tthe Free Software Foundation; either version 2, or (at your option) any",
	"\tlater version. A copy of that license is included in the file COPYING",
	"\tincluded with the lptops distribution. The license is also available",
	"\tfrom the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139,",
	"\tUSA, and at its World-Wide Web site at",
	"",          
	"\t\thttp://www.gnu.org/licenses/",
	CNULL
    };

    Out_Lines(stderr,copyright);
    exit(EXIT_SUCCESS);
}

static void
#if STDC
Opt_Error(
const char			*fmt,
const char			*option
)
#else /* NOT STDC */
Opt_Error(fmt,option)
const char			*fmt;
const char			*option;
#endif /* STDC */
{
    Warning((fmt == CNULL) ? "Unrecognized option [%s]" : fmt,option);
    bad_option = TRUE;
}

static void
#if STDC
Opt_Font(
const char			*option,
const char			*value
)
#else /* NOT STDC */
Opt_Font(option,value)
const char			*option;
const char			*value;
#endif
{
    int				(*compare) ARGS((const char*, const char*));
    int				j;	/* loop index */
    int				k;	/* loop index */
    int				m;	/* loop index */
    int				n;	/* loop index */
    const char			*name[5*4]; /* 5 families, 4 names of each */

    /*******************************************************************

    The font argument is either a single name, to select a font family,
    or else a list of 5*n names of the form

    fullname-normal,alias,alias,alias, fullname-bold,alias,alias,alias,
    fullname-italic,alias,alias,alias, fullname-bolditalic,alias,alias,alias,
    fullname-typewriter,alias,alias,alias,

    with 0 through 3 aliases to define a new font family.  For example,

    Times-Roman,T,R,ti,  Times-Bold,TB,B,Ti,  Times-Italic,TI,I,tI,
    Times-BoldItalic,TBI,BI,TI,  Courier,C,C,C

    would define the built-in Times family with its full PostScript
    names, plus the short, ditroff, and TranScript aliases.

    Another example, without aliases, is

    UtopiaRegular,  UtopiaBold,  UtopiaItalic,  UtopiaBoldItalic, Courier

    and with one alias,

    UtopiaRegular,UR,  UtopiaBold,UB,  UtopiaItalic,UI,  UtopiaBoldItalic,UBI,
    Courier, C

    *******************************************************************/

    n = Tokenize(value,name,sizeof(name)/sizeof(name[0]),FONTSEPCHARS);

    if (n > 1)				/* then expect font family definition */
    {
	if ((n % 5) != 0)
	{
	    Opt_Error("Invalid font family [%s]: need 5*n names",option);
	    return;
	}
	m = n / 5;			/* number of names per ps_xxx[] entry */
	for (k = 0; (k < (MAXFONTS-8)) && (ps_normal[k] != CNULL);
	    k += 4)
	    /* NO-OP */;		/* find first empty font slot */
	if (k >= MAXFONTS)
	{
	    Opt_Error("Too many font families at option [%s]",option);
	    return;
	}
	for (j = 0; j < m; ++j)		/* save user-specified aliases */
	{
	    ps_normal	[k + j] = name[j + 0*m];
	    ps_bold	[k + j] = name[j + 1*m];
	    ps_italic	[k + j] = name[j + 2*m];
	    ps_both	[k + j] = name[j + 3*m];
	    ps_typewriter[k + j] = name[j + 4*m];
	}
	for (j = m; j < 4; ++j) /* remaining aliases are empty strings */
	{
	    ps_normal	[k + j] = "";
	    ps_bold	[k + j] = "";
	    ps_italic	[k + j] = "";
	    ps_both	[k + j] = "";
	    ps_typewriter[k + j] = "";
	}
	ps_normal	[k + 4] = CNULL;
	ps_bold		[k + 4] = CNULL;
	ps_italic	[k + 4] = CNULL;
	ps_both		[k + 4] = CNULL;
	ps_typewriter	[k + 4] = CNULL;
	return;
    }
    else if (n == 1)			/* otherwise, have a single font name */
    {
	normal.name = name[0];		/* remember the name globally */
	for (n = 0; ps_normal[n] != CNULL; n += 4)
	{
	    for (m = 3; m >= 0; --m)
	    {
		/* compare case-sensitive ditroff and TranScript
		   names before case-insensitive ones */
		compare = ((m > 1) ? (COMPARE_CAST)strcmp : (COMPARE_CAST)stricmp);
		if ((*compare)(normal.name,ps_normal[n+m]) == 0)
		{
		    Set_Family(ps_normal[n], (n <= PS_RESIDENT),
			ps_bold[n], (n <= PS_RESIDENT),
			ps_italic[n], (n <= PS_RESIDENT),
			ps_both[n], (n <= PS_RESIDENT),
			ps_typewriter[n], (n <= PS_RESIDENT));
		    return;
		}
		else if ((*compare)(normal.name,ps_bold[n+m]) == 0)
		{
		    Set_Family(ps_bold[n], (n <= PS_RESIDENT),
			ps_normal[n], (n <= PS_RESIDENT),
			ps_italic[n], (n <= PS_RESIDENT),
			ps_both[n], (n <= PS_RESIDENT),
			ps_typewriter[n], (n <= PS_RESIDENT));
		    return;
		}
		else if ((*compare)(normal.name,ps_italic[n+m]) == 0)
		{
		    Set_Family(ps_italic[n], (n <= PS_RESIDENT),
			ps_both[n], (n <= PS_RESIDENT),
			ps_normal[n], (n <= PS_RESIDENT),
			ps_both[n], (n <= PS_RESIDENT),
			ps_typewriter[n], (n <= PS_RESIDENT));
		    return;
		}
		else if ((*compare)(normal.name,ps_both[n+m]) == 0)
		{
		    Set_Family(ps_both[n], (n <= PS_RESIDENT),
			ps_normal[n], (n <= PS_RESIDENT),
			ps_italic[n], (n <= PS_RESIDENT),
			ps_both[n], (n <= PS_RESIDENT),
			ps_typewriter[n], (n <= PS_RESIDENT));
		    return;
		}
		else if ((*compare)(normal.name,ps_typewriter[n+m]) == 0)
		{
		    Set_Family(ps_typewriter[n], (n <= PS_RESIDENT),
			ps_bold[n], (n <= PS_RESIDENT),
			ps_italic[n], (n <= PS_RESIDENT),
			ps_both[n], (n <= PS_RESIDENT),
			ps_typewriter[n], (n <= PS_RESIDENT));
		    return;
		}
	    }
	}
    }
    Opt_Error("Illegal font name [%s].  Possible values are:",
	      (name[0] == CNULL) ? "" : name[0]);
    Dump_Font("Normal fonts",ps_normal);
    Dump_Font("Bold fonts",ps_bold);
    Dump_Font("Oblique/Italic fonts",ps_italic);
    Dump_Font("BoldOblique/BoldItalic fonts",ps_both);
    Dump_Font("Typewriter fonts",ps_typewriter);
}

static void
#if STDC
Opt_Help(
const char			*option,
const char			*value
)
#else /* NOT STDC */
Opt_Help(option,value)
const char			*option; /* NOT USED */
const char			*value;	/* NOT USED */
#endif /* STDC */
{
    Usage(EXIT_SUCCESS);
}

static void
#if STDC
Opt_Page(
const char			*option,
const char			*value
)
#else /* NOT STDC */
Opt_Page(option,value)			/* value -> "Nxxxnnn" */
const char			*option;
const char			*value;
#endif /* STDC */
{
    do_page_number = TRUE;
    init_page_number = 1;
    page_flag = (PAGE_T | PAGE_C);    /* default location */
    for ( ; *value; ++value)
    {
	switch(*value)
	{
	case 't':
	case 'T':
	    page_flag &= ~(PAGE_B | PAGE_T);
	    page_flag |= PAGE_T;
	    break;

	case 'b':
	case 'B':
	    page_flag &= ~(PAGE_B | PAGE_T);
	    page_flag |= PAGE_B;
	    break;

	case 'l':
	case 'L':
	    page_flag &= ~(PAGE_L | PAGE_C | PAGE_R);
	    page_flag |= PAGE_L;
	    break;

	case 'c':
	case 'C':
	    page_flag &= ~(PAGE_L | PAGE_C | PAGE_R);
	    page_flag |= PAGE_C;
	    break;

	case 'r':
	case 'R':
	    page_flag &= ~(PAGE_L | PAGE_C | PAGE_R);
	    page_flag |= PAGE_R;
	    break;

	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	    init_page_number = (int)strtol(value,(char**)NULL,0);
	    while (isdigit(*value))
		++value;
	    if (*value == NUL)
		break;
			    /* else fall through to error exit */
	default:
	    Opt_Error("Unrecognized page number option [%s]",option);
	}
	page_number = init_page_number;	/* this needs to be reset */
	if (*value == NUL)
	    break;
    }
}

static void
#if STDC
Opt_Paper(
const char			*option,
const char			*value
)
#else /* NOT STDC */
Opt_Paper(option,value)
const char			*option;
const char			*value;
#endif /* STDC */
{
    BOOLEAN			bad;
    int				k;
    int				n;
    const char			*name[8];
    double			h;
    double			w;
    double			left;
    double			right;
    double			bottom;
    double			top;

    errno = 0;				/* clear global error flag */
    n = Tokenize(value,name,sizeof(name)/sizeof(name[0]),PAPERSEPCHARS);

    if (n == 1)					  /* look for existing paper */
    {
	for (k = 0; paper[k].name != CNULL; ++k)
	{
	    if (stricmp(paper[k].name,name[0]) == 0)
	    {
		paper_name	= paper[k].name;
		page_width	= DIMEN_TO_QSU(paper[k].width);
		page_height	= DIMEN_TO_QSU(paper[k].height);
		landscape_mode	= (page_height < page_width);
		left_margin	= DIMEN_TO_QSU(paper[k].left_margin);
		right_margin	= DIMEN_TO_QSU(paper[k].right_margin);
		bottom_margin	= DIMEN_TO_QSU(paper[k].bottom_margin);
		top_margin	= DIMEN_TO_QSU(paper[k].top_margin);
		break;
	    }
	}
	if (paper[k].name == CNULL)
	{
	    Opt_Error("Unknown paper name in option [%s]",option);
	    Dump_Paper();
	}
    }
    else if (n >= 3)				  /* define new paper form */
    {
	for (k = n; k < 7; ++k)		/* supply defaults for missing values */
	    name[k] = "1in";

	bad = FALSE;

	w = inch(name[1]);
	bad = bad || (w < 1.0) || (w > 54.0);

	h = inch(name[2]);
	bad = bad || (h < 1.0) || (h > 54.0);

	left = inch(name[3]);
	bad = bad || (left < 0.0) || (left > w);

	right = inch(name[4]);
	bad = bad || (right < 0.0) || (right > w);

	bottom = inch(name[5]);
	bad = bad || (bottom < 0.0) || (bottom > h);

	top = inch(name[6]);
	bad = bad || (top < 0.0) || (top > h);

	if (bad)
	    Opt_Error("Bad paper definition in [%s]",option);
	else
	{
	    for (k = 0; k < (MAXPAPER-1); ++k)
	    {				  /* find next empty slot in table */
		if (paper[k].name == CNULL)
		    break;
	    }
	    if (k < (MAXPAPER-1))	/* then have space for new entry */
	    {
		paper[k].name		= name[0];
		paper[k].width		= name[1];
		paper[k].height		= name[2];
		paper[k].left_margin	= name[3];
		paper[k].right_margin	= name[4];
		paper[k].bottom_margin	= name[5];
		paper[k].top_margin	= name[6];

		paper[k+1].name		= CNULL;
		paper[k+1].width	= CNULL;
		paper[k+1].height	= CNULL;
		paper[k+1].left_margin	= CNULL;
		paper[k+1].right_margin	= CNULL;
		paper[k+1].bottom_margin= CNULL;
		paper[k+1].top_margin	= CNULL;
	    }
	    else
		Opt_Error("Too many paper definitions at [%s]",option);
	}
    }
    else
	Opt_Error("Bad paper definition [%s]",option);
}

static void
#if STDC
Opt_SelectPages(
const char			*option,
const char			*value
)
#else /* NOT STDC */
Opt_SelectPages(option,value)
const char			*option;
const char			*value;
#endif /* STDC */
{
    int				nbytes;
    int				nitems;
    const char			*p;
    char			*save;
    char			*token;

    for (p = value, nitems = 1; *p; ++p)
    {
	if (*p == ',')
	    nitems++;	/* count number of comma-separated page ranges */
    }

    nbytes = (npages + nitems) * (int)sizeof(PAGE_SELECTION);
    pages = (pages == (PAGE_SELECTION*)NULL)
	? (PAGE_SELECTION*)malloc(nbytes)
	: (PAGE_SELECTION*)REALLOC(pages, nbytes);

    save = Strdup(value);

    for (token = strtok(save,",");
	 (token != (char*)NULL);
	  token = strtok((char*)NULL,","))
    {
	nitems = tscanf(token,
			"%d%*[^,0-9]%d%*[^,0-9]%d",
			&pages[npages].begin,
			&pages[npages].end,
			&pages[npages].step);
	switch (nitems)
	{
	case 1:
	    pages[npages].end = pages[npages].begin;
	    pages[npages].step = 1;
	    break;

	case 2:
	    pages[npages].step = 1;
	    break;

	case 3:
	    pages[npages].step = MAX(1,pages[npages].step);
	    break;

	default:
	    Opt_Error("Bad page range: [%s]",token);
	    return;
	}
	npages++;
    }

#if 0
    {
	int k;
	for (k = 0; k < npages; ++k)
	    FPRINTF(stderr,"Page selection: %ld--%ld by %ld\n",
		    pages[k].begin,
		    pages[k].end,
		    pages[k].step);
    }
#endif

    FREE(save);
}

static void
#if STDC
Opt_Version(
const char			*option,
const char			*value
)
#else /* NOT STDC */
Opt_Version(option,value)
const char			*option;
const char			*value;
#endif /* STDC */
{
    static const char *version[] =
    {
	VERSION,
	CNULL
    };

    Out_Lines(stderr,version);
    exit(EXIT_SUCCESS);
}

static void
#if STDC
Out_Atsign(
FILE *fp
)
#else /* NOT STDC */
Out_Atsign(fp)
FILE *fp;
#endif /* STDC */
{
    register int		c;	/* character */

    if (scribe)
    {
	c = inchar(fp);		/* peek ahead to next character */
	switch (c)
	{
	case 'b':		/* expect @b[...] font change */
	case 'B':
	case 'h':		/* expect @h[...] font change */
	case 'H':
	case 'i':		/* expect @i[...] font change */
	case 'I':
	case 't':		/* expect @t[...] font change */
	case 'T':
	case 'u':		/* expect @u[...] font change */
	case 'U':		/* (underline font is extension to Scribe) */
	    Do_Scribe_string(fp,islower(c) ? toupper(c) : c);
	    break;

	case '@':		/* @@ means output single @ */
	    Out_Other('@');
	    break;

	default:
	    UNGETCHAR(c);	/* put back the peeked-at character, */
	    Out_Other('@');
	    break;
	}
    }
    else			/* not scribe */
	Out_Other('@');
}

static void
Out_Backspace(VOID_ARG)
{
    if (in_col > 0)
    {
	in_col--;		/* cannot sensibly update line_width */
	OUTSTRCHAR('\\');
	OUTSTRCHAR('b');
    }
}

static void
Out_Bullet(VOID_ARG)
{
    End_Text();
    OUTSTR(" D");
    in_col++;
    line_width += BULLETWIDTH;	/* fixed by choice of Symbol font */
}

static void
#if STDC
Out_Char(
int				c
)
#else /* NOT STDC */
Out_Char(c)
int				c;
#endif /* STDC */
{
    /* All PostScript output comes through this function, so that we
       can track the output line widths, and generate newlines as
       needed.  We can safely wrap long lines only at a discardable
       space, provided we observe the presence of comments, most of
       which cannot be wrapped, and characters strings, which require
       backslash-newline continuation.  String wrapping is handled
       separately by Out_StrChar().  It doesn't matter if we
       occasionally falsely misidentify a string with a percent as a
       comment; it is only crucial that we not fail to recognize true
       comments.  */

    if ((c == '\n') ||
	((out_column >= MAXLINEWIDTH) && !comment_line && isspace(c)))
    {
	if (output_this_page)
	    (void)putc((char)'\n',tf);
	out_column = 0;
	comment_line = FALSE;
    }
    else
    {
	if (c == '%')
	    comment_line = TRUE;
	if (output_this_page)
	    (void)putc((char)c,tf);
	if (c == '\t')
	    out_column = (out_column + 7) & ~07;
	else
	    out_column++;
    }
}

static void
#if STDC
Out_Escape(
FILE *fp
)
#else /* NOT STDC */
Out_Escape(fp)
FILE *fp;
#endif /* STDC */
{
    int				c;

    if (manpage)
    {
	c = inchar(fp);			/* peek ahead one character */
	switch (c)
	{
	case '7':
	    Out_Vskip(1.0);		/* reverse line feed */
	    break;

	case '8':
	    Out_Vskip(0.5);		/* reverse half line feed */
	    break;

	case '9':
	    Out_Vskip(-0.5);		/* half line feed */
	    break;

	default:
	    UNGETCHAR(c);
	    Out_Other('\033');
	    break;
	}
    }
    else
	Out_Other('\033');
}

static void
#if STDC
Out_Fontdef(
const char *macro,
const char *fontname
)
#else /* NOT STDC */
Out_Fontdef(macro,fontname)
const char *macro;
const char *fontname;
#endif /* STDC */
{
    FPRINTF(tf, "/F%s { %% F%s -- select %s font", macro, macro, fontname);
    OUTNL();

    FPRINTF(tf, "\t/%s findfont", fontname);
    OUTNL();

    if (ISO8859_1_charset)
    {
	FPRINTF(tf, "\tfindISO {reencodeISO /%s-ISO exch definefont} if",
	    fontname);
	OUTNL();
    }

    FPRINTF(tf, "\tPOINTSIZE scalefont setfont");
    OUTNL();

    FPRINTF(tf, "} def");
    OUTNL();

    FPRINTF(tf,
	"/%s { F%s show } def %% (string) %s -- show string in %s font",
	macro, macro, macro, fontname);
    OUTNL();
}

static void
#if STDC
Out_Font_Resource(
const char			*name,
int				nfonts
)
#else /* NOT STDC */
Out_Font_Resource(name,nfonts)
const char			*name;
int				nfonts;
#endif /* STDC */
{
    if (nfonts > 0)
    {
	OUTNL();
	OUTSTR("%%+\t");
    }
    OUTSTR("font ");
    OUTSTR(name);
}

static void
Out_Formfeed(VOID_ARG)
{
    End_Line();
    in_col = 0;
    line_width = 0L;
    y = BOT_Y - 1;	/* force end-of-column action next time around */
}

static void
#if STDC
Out_Lines(
FILE *fp,
const char *lines[]
)
#else /* NOT STDC */
Out_Lines(fp,lines)
FILE *fp;
const char *lines[];
#endif /* STDC */
{
    int				k; /* index into lines[] */

    if (output_this_page)
    {
	for (k = 0; lines[k] != CNULL; ++k)
	{
	    fputs(lines[k],fp);
	    NEWLINE(fp);
	}
    }
}

static int
#if STDC
Out_Manchar(
FILE *fp,
int c
)
#else /* NOT STDC */
Out_Manchar(fp,c)		/* return c, or EOF if c was output */
FILE *fp;
int c;
#endif /* STDC */
{
    int				next_c;
    int				next_next_c;
    Font			*save_font;
    int				save_id;

    /* non-printing chars and underscore never overprint */
    if ((c <= (int)' ') || (c == (int)'_'))
	return (c);

    /* Output <char>(<backspace><char>)+ as (<char>)B for boldface type */

    next_c = inchar(fp);	/* peek ahead one character */
    if (next_c != (int)'\b')	/* if not a backspace, then we are done */
    {
	UNGETCHAR(next_c);
	return (c);
    }

    next_next_c = inchar(fp);	/* peek ahead another character */

    if (   ((c == '+') && (next_next_c == 'o'))	 /* <+><BS><o> -> bullet */
	|| ((c == 'o') && (next_next_c == '+')) ) /* <o><BS><+> -> bullet */
    {					/* SunOS uses one, Solaris the other */
	Out_Bullet();
	return (EOF);
    }
    else if (c != next_next_c)	/* not overstruck sequence */
    {
	UNGETCHAR(next_next_c); /* put back two-character lookahead */
	UNGETCHAR(next_c);
	return (c);
    }

    UNGETCHAR(next_next_c);	/* put back three-character lookahead */
    UNGETCHAR(next_c);
    UNGETCHAR(c);

    save_id = current_id;
    save_font = current_font;
    Beg_Text();
    current_id = (int)'B';
    current_font = &bold;
    while ((c = Get_Bold_Char(fp)) != EOF)
	Out_Other(c);
    End_Text();
    current_id = save_id;
    current_font = save_font;

    return (EOF);
}

static void
#if STDC
Out_Need(
int				count
)
#else /* NOT STDC */
Out_Need(count)
int				count;
#endif /* STDC */
{
    /* When a multicharacter sequence inside a parenthesized
       PostScript string must be output without the possibility of a
       line continuation, we increase out_column to force generation
       of a backslash-newline in Out_StrChar(). */

    if ((out_column + count) >= MAXLINEWIDTH)
	out_column = MAXLINEWIDTH;
}

static void
Out_Newline(VOID_ARG)
{
    if (BOL())
	Beg_Text();		/* force empty string if necessary */
    End_Line();
    in_col = 0;
    line_number++;
    line_width = 0L;
    y -= baseline_skip;
}

static void
#if STDC
Out_Other(int c)
#else
Out_Other(c)
int c;
#endif
{ /* output a string character, WITH quoting, width and column update, and */
  /* line wrap check */

    Wrap_Check(c);
    if ((c < 040) || (0176 < c))
    {
	Out_Need(4);
	OUTSTRCHAR('\\');
	OUTSTRCHAR((int)'0' + ((c >> 6) & 03));
	OUTSTRCHAR((int)'0' + ((c >> 3) & 07));
	OUTSTRCHAR((int)'0' + ((c >> 0) & 07));
    }
    else if ((c == (int)'\\') || (c == (int)'(') || (c == (int)')'))
    {
	Out_Need(2);
	OUTSTRCHAR('\\');
	OUTSTRCHAR(c);
    }
    else
	OUTSTRCHAR(c);
    line_width += CHARWIDTH(c);
    in_col++;
}

static void
#if STDC
Out_Return(
FILE *fp
)
#else /* NOT STDC */
Out_Return(fp)
FILE *fp;
#endif /* STDC */
{
    register int		c;	/* character */
				/* check for special case of overprinting */
    if ((c = inchar(fp)) != (int)'\n') /* peek ahead one character */
    {				/* and if not LF, then have overstrike */
	End_String();
	if (in_col > 0)
	{
	    in_col = 0;
	    line_width = 0L;
	    OUTCHAR('O');
	    OUTNL();		/* this lines up under previous line */
	}
    }
				/* else do nothing, reducing CR LF to LF */
    UNGETCHAR(c);  /* put back the peeked-at character */
}

static void
#if STDC
Out_StrChar(
int c
)
#else /* NOT STDC */
Out_StrChar(c)
int c;
#endif /* STDC */
{	/* output c in a parenthesized string, WITHOUT special quoting */

    if (level == 0) /* if not in a parenthesized string yet, start new one */
	Beg_String();
    if (out_column >= (MAXLINEWIDTH-1))
    {				/* wrap long string with backslash-newline */
	OUTCHAR('\\');
	OUTNL();
    }
    OUTCHAR(c);
    in_str++;
}

static void
#if STDC
Out_String(
const char			*s
)
#else /* NOT STDC */
Out_String(s)
const char			*s;
#endif /* STDC */
{
    if (!comment_line && (((int)strlen(s) + out_column) > (MAXLINEWIDTH - 1)))
	OUTNL();
    for ( ; *s; ++s)
	Out_Char(*s);
}

static void
Out_Tab(VOID_ARG)
{
    do
    {
	Out_Other(' ');
    }
    while (in_col & 07); /* blank fill to multiple of 8 columns */
}

static void
#if STDC
Out_Underline(
FILE *fp
)
#else /* NOT STDC */
Out_Underline(fp)
FILE *fp;
#endif /* STDC */
{
    register int		c;	/* character */
    Font			*save_font;
    int				save_id;

    /* check for special case of (<underline><BS><char>)+ */

    if (manpage)
    {
	c = inchar(fp);			/* peek ahead one character */
	if (c == (int)'\b')		/* then have _<BS><char> */
	{
	    UNGETCHAR(c);
	    UNGETCHAR('_');
	    save_id = current_id;
	    save_font = current_font;
	    Beg_Text();
	    current_id = (int)'U';
	    current_font = &normal;
	    while ((c = Get_Underlined_Char(fp)) != EOF)
		Out_Other(c);
	    End_Text();
	    current_id = save_id;
	    current_font = save_font;
	}
	else
	{
	    UNGETCHAR(c);		/* put back the lookahead */
	    Out_Other('_');
	}
    }
    else				/* just output literal underline */
	Out_Other('_');
}

static void
#if STDC
Out_Vskip(
double				v
)
#else /* NOT STDC */
Out_Vskip(v)
double				v;
#endif /* STDC */
{
    char			d[25];

    End_Text();
    y += (COORDINATE)(v * (double)baseline_skip);
    if (v == 0.5)			/* eqn superscript */
	OUTSTR(" u ");
    else if (v == -0.5)			/* eqn subscript */
	OUTSTR(" d ");
    else				/* arbitrary vskip */
    {
	SPRINTF(d," %g",v);
	OUTSTR(d);
	OUTSTR(" V ");
    }
}

static size_t
#if STDC
PFB_Length(
FILE				*fpin
)
#else
PFB_Length(fpin)
FILE				*fpin;
#endif
{
    size_t			n;

    n  = (size_t)((fgetc(fpin) & 0xff));
    n |= (size_t)((fgetc(fpin) & 0xff) << 8);
    n |= (size_t)((fgetc(fpin) & 0xff) << 16);
    n |= (size_t)((fgetc(fpin) & 0xff) << 24);

    return (n);
}

static void
#if STDC
Read_Metrics(
Font				*font
)
#else /* NOT STDC */
Read_Metrics(font)
Font				*font;
#endif /* STDC */
{
    char			afmbuf[BUFSIZ];
    FILE			*afmfile; /* Adobe Font Metric file handle */
    int				cpos;
    int				charwidth;
    int				dataflag;
    char			filename[MAXPATHLEN+1];
    char			*fontchar;
    const char			*fontfile;
    COORDINATE			width;

#ifdef DEBUG
    FPRINTF(stderr, "DEBUG: point_size %d", point_size);
    NEWLINE(stderr);
#endif

    SPRINTF(filename,"%s%s",font->name,AFMEXT);
    fontfile = findfile(afmpath, filename);

    if ((fontfile == CNULL)) /* then try the map file */
	fontfile = Map_Font(font->name,AFMEXT);

    if ((fontfile == CNULL) ||
	(afmfile = fopen(fontfile, FOPEN_R)) == (FILE*)NULL)
    {
	if (!quiet)
	    Warning("Cannot open Adobe Font Metric file %s: defaulting to \
11bp-Courier", (fontfile == CNULL) ? filename : fontfile);
	font->name = ps_normal[PS_DEFAULT];
	font->resident = (PS_DEFAULT <= PS_RESIDENT);
	point_size = DIMEN_TO_QSU("11bp");
	width = (6L * point_size) / 10L; /* Courier width = 0.6*height */
	for (cpos = 0; cpos < MAXFONTSIZE; cpos++)
	    font->CharWidth[cpos] = width;
	return;
    }
    else
    {
	for (cpos = 0; cpos < MAXFONTSIZE; cpos++)
	    font->CharWidth[cpos] = 0L;
    }

    dataflag = 0;
    while (fgets(afmbuf, sizeof(afmbuf), afmfile) != NULL)
    {
	if ((fontchar = strchr(afmbuf, '\n')) == 0)
	    Fatal("AFM file %s line too long", fontfile);
	*fontchar = NUL;
	if (*afmbuf == NUL)
	    continue;
	else if (strncmp(afmbuf, "StartCharMetrics", 16) == 0)
	{
	    dataflag++;
	    continue;
	}
	else if (strcmp(afmbuf, "EndCharMetrics") == 0)
	    break;		/* EndCharMetrics tells us the data is read */
	else if (dataflag)
	{
	    /* read the char position and its width */
	    if (tscanf(afmbuf, "C %d ; WX %d ;", &cpos, &charwidth) != 2)
		Fatal("Bad data in AFM file %s", fontfile);
	    else if (cpos == -1) /* then we have seen unencoded character */
		break;		/* so stop reading .afm file */
	    else if ((0 <= cpos) && (cpos < MAXFONTSIZE))
		font->CharWidth[cpos] = (COORDINATE)(((double)charwidth *
		    (double)point_size) / 1000.0);
	}
    }				/* end while */

#ifdef DEBUG
    for (cpos = 0; cpos < MAXFONTSIZE; cpos++)
    {
	FPRINTF(stderr, "DEBUG: char %c %d %d", cpos, cpos,
	    font->CharWidth[cpos]);
	NEWLINE(stderr);
    }
    (void)fflush(stderr);
#endif

    (void)fclose(afmfile);
    errno = 0;				/* clear global error flag */
}

static void
Reencode()
{
    static const char *reencoding[] =
    {
	"/reencodeISO {",
	"\tdup length dict begin",
	"\t\t{1 index /FID ne {def}{pop pop} ifelse} forall",
	"\t\t/Encoding ISOLatin1Encoding def",
	"\t\tcurrentdict",
	"\tend",
	"} def",
	"/findISO {",
	"\tdup /FontType known {",
	"\t\tdup /FontType get 3 ne {",
	"\t\t\tdup /CharStrings known {",
	"\t\t\t\tdup /CharStrings get /Thorn known {",
	"\t\t\t\t\ttrue",
	"\t\t\t\t}{ false } ifelse",
	"\t\t\t}{ false } ifelse",
	"\t\t}{ false } ifelse",
	"\t}{ false } ifelse",
	"} def",
	CNULL
    };

    Out_Lines(tf,reencoding);
}

static void
Reverse_Pages(VOID_ARG)
{
    FILE			*fpin;	/* input file */
    FILE			*fpout;	/* output file */
    register int		k;	/* loop index */
    long			last;
    long			next;	/* result of fseek() for */
					/* page_table[] construction */

#if OS_UNIX
    rewind(tf);		/* file was opened in read/write (update) mode */
#else
    /* Now close the temporary file and reopen for input */

    if (fclose(tf))
	Fatal("Close failure on temporary file [%s]", TEMPFILE_1);

    if ((tf = Fopen_Tempfile(TEMPFILE_1,FOPEN_R)) == (FILE *)NULL)
	Fatal("Cannot reopen temporary file [%s]", TEMPFILE_1);
#endif

    /* On HP-UX, ftell(stdout) fails if stdout is a pipe, even though */
    /* all we really want to know is how many bytes have been written, */
    /* and that information ought to be available.  The 1989 ANSI/ISO C */
    /* Standard allows this behavior, so we provide a workaround of */
    /* reversing pages into a second temporary file, then copy that file */
    /* to stdout.  This is only done if ftell(stdout) is found to fail. */

    errno = 0;
    if ((ftell(stdout) < 0) || (errno != 0))
    {	/* cannot ftell() on stdout, so copy via intermediate file */
	if ((fpout = Fopen_Tempfile(TEMPFILE_2,FOPEN_RW)) == (FILE *)NULL)
	    Fatal("Cannot open temporary file [%s]", TEMPFILE_2);

#if OS_UNIX
	unlink(TEMPFILE_2); /* so temp file disappears even if we abort */
#endif

    }
    else
	fpout = stdout;
    fpin = tf;

    /* Transfer complete pages in reverse order to fpout using large blocks */

    if (!quiet)
    {
	NEWLINE(stderr);
	FPRINTF(stderr,"[Reversing pages");
    }
    Copy_Block(fpin,fpout,0L,page_table[0]-1);
					/* first page is file header stuff */
    last = page_table[MIN(MAXPAGE,page_number-init_page_number+1)];
    for (k = MIN(MAXPAGE,page_number-init_page_number+1)-1; k >= 0; --k)
    {					/* copy remainder in reverse order */
	Check_Page_Selection(k + init_page_number);
	if (!quiet && output_this_page)
	    (void)putc('.',stderr);
	next = (long)ftell(fpout);
	Copy_Block(fpin,fpout,page_table[k],last-1);
	last = page_table[k];
	page_table[k] = next;
    }
    if (fpout != stdout)
    {
	last = ftell(fpout);
	rewind(fpout);
	Copy_Block(fpout,stdout,0L,last-1);

#if OS_UNIX
					/* file was unlinked at open time */
#else
	remove(TEMPFILE_2);		/* delete the temporary file */
#endif

    }

    if (!quiet)
	(void)putc(']',stderr);

    if (fclose(tf))
	Fatal("Close failure on temporary file [%s]", TEMPFILE_1);

#if OS_UNIX
					/* file was unlinked at open time */
#else
    remove(TEMPFILE_1);			/* delete the temporary file */
#endif
}

static void
#if STDC
Set_Family(
const char			*name_normal,
BOOLEAN				resident_normal,
const char			*name_bold,
BOOLEAN				resident_bold,
const char			*name_italic,
BOOLEAN				resident_italic,
const char			*name_both,
BOOLEAN				resident_both,
const char			*name_typewriter,
BOOLEAN				resident_typewriter
)
#else /* NOT STDC */
Set_Family(name_normal, resident_normal, name_bold, resident_bold, name_italic,
    resident_italic, name_both, resident_both,
    name_typewriter, resident_typewriter)
const char			*name_normal;
BOOLEAN				resident_normal;
const char			*name_bold;
BOOLEAN				resident_bold;
const char			*name_italic;
BOOLEAN				resident_italic;
const char			*name_both;
BOOLEAN				resident_both;
const char			*name_typewriter;
BOOLEAN				resident_typewriter;
#endif /* STDC */
{
    bold.name = name_bold;
    bold.resident = resident_bold;

    bolditalic.name = name_both;
    bolditalic.resident = resident_both;

    italic.name = name_italic;
    italic.resident = resident_italic;

    normal.name = name_normal;
    normal.resident = resident_normal;

    typewriter.name = name_typewriter;
    typewriter.resident = resident_typewriter;
}

char*
#if STDC
Strdup(
const char			*s
)
#else /* NOT STDC */
Strdup(s)
const char			*s;
#endif /* STDC */
{
    char			*p;

    p = (char*)malloc((size_t)(strlen(s) + 1));
    if (p == (char*)NULL)
	Fatal("Out of memory","");
    (void)strcpy(p,s);
    return (p);
}

static void
#if STDC
Strip_Comments(
char				line[],
int				comment_char
)
#else /* NOT STDC */
Strip_Comments(line,comment_char)
char				line[];
int				comment_char;
#endif /* STDC */
{
    char			*p;	/* pointer into line[] */

    for (p = strchr(line,NUL) - 1; isspace(*p); --p)
	*p = NUL;			/* trim trailing white space */

    /* strip comments from comment_char to end-of-line, ignoring */
    /* backslash escaped comment characters */

    for (p = strchr(line,comment_char); (p != (char*)NULL);
	 p = strchr(p+1,comment_char))
    {
	if ( (p == &line[0]) || ((p > &line[0]) && (*(p-1) != '\\')) )
	{				/* then found a real comment */
	    *p = NUL;			/* so discard rest of line */
	    break;
	}
    }
    for (p = strchr(line,NUL) - 1; (p >= &line[0]) && isspace(*p); --p)
	*p = NUL;			/* trim trailing space */
}

static int
#if STDC
Tokenize(
const char			*s,
const char			*name[],
int				maxname,
const char			*sepchars
)
#else /* NOT STDC */
Tokenize(s,name,maxname,sepchars)
const char			*s;
const char			*name[];
int				maxname;
const char			*sepchars;
#endif /* STDC */
{
    int				n;
    char 			*p;

    p = Strdup(s);			/* get modifiable copy */
    name[maxname-1]= CNULL;		/* list terminator */
    for (n = 0; n < (maxname-1); ++n)
    {					/* split argument into separate names */
	name[n] = (const char*)strtok(p,sepchars);
	if (name[n] == CNULL)
	    break;
	p = (char*)NULL;
    }
    return (n);
}

static int
#if STDC
ungetchar(
int c
)
#else /* NOT STDC */
ungetchar(c)		/* unget c, up to MAXPUSHBACK chars, return c */
int c;			/* NB: only ONE (unnamed) input file is supported */
#endif /* STDC */
{
    if (npushback < MAXPUSHBACK)
	pushback[npushback++] = c;
    else
	Fatal("Internal error: pushback buffer overflow",CNULL);
    return (c);
}

static void
#if STDC
Usage(
int exit_status
)
#else
Usage(exit_status)
int exit_status;
#endif
{
    static const char *message[] =
    {
	"-? -author -bottom=# -cn -clip[=n] -copyleft -copyright -e[=n] \
-font=fontname",
	"\t-font=fontfamilylist -help -i=# -landscape[=n] -left=# -m[=n] \
-manpage[=n]",
	"\t-n[=[tb][lcr][n]] -o[=#] -p=# -page=#-#,#-#,...,#-# -paper=name",
	"\t-paper=name,width,height[,left[,right[,bottom[,top]]]] -quiet[=n]",
	"\t-reverse[=n] -right=# -s[=n] -scribe[=n] -top=# -version -wrap[=n] \
-y=# ",
	"\t[filelist] >outfile",
	"",
	"Default options:",
	"\t-bottom=1.0in -c=1 -font=Courier -left=1.0in -m=1 -p=10bp -paper=A",
	"\t-right=1.0in -s=0 -top=1.0in -y=12bp",
	"",
	"Command-line options may begin with -- as well as -.",
	CNULL
    };

    NEWLINE(stderr);
    FPRINTF(stderr,"Usage: %s ",program_name); /* NO newline here! */
    Out_Lines(stderr,message);
    exit(exit_status);
}

static void
#if STDC
Warning(
const char *fmt,
const char *arg
)
#else /* NOT STDC */
Warning(fmt,arg)
const char *fmt;
const char *arg;
#endif
{
    NEWLINE(stderr);		/* ensure start of new line */
    FPRINTF(stderr, "%s%s: ", msg_prefix, program_name);
    FPRINTF(stderr, fmt, arg);
    NEWLINE(stderr);
    if (errno)
    {
	FPRINTF(stderr, "%s%s: ", msg_prefix, program_name);
	(void)perror("perror() says");
    }
}

static void
#if STDC
Wrap_Check(
int c
)
#else /* NOT STDC */
Wrap_Check(c)
int c;
#endif /* STDC */
{
    if (wrap && (c > ' ') && ((line_width + CHARWIDTH(c)) > x_right))
	Out_Newline();	/* visible chars can cause line wraps */
}
